"""Classification trainer/evaluator implementation module."""
import logging
from typing import AnyStr  # noqa: F401

import torch
import torch.optim

import thelper.concepts
import thelper.typedefs as typ  # noqa: F401
import thelper.utils
from thelper.train.base import Trainer

logger = logging.getLogger(__name__)


@thelper.concepts.classification
class ImageClassifTrainer(Trainer):
    """Trainer interface specialized for image classification.

    This class implements the abstract functions of :class:`thelper.train.base.Trainer` required to train/evaluate
    a model for image classification or recognition. It also provides a utility function for fetching i/o packets
    (images, class labels) from a sample, and that converts those into tensors for forwarding and loss estimation.

    .. seealso::
        | :class:`thelper.train.base.Trainer`
    """

    def __init__(self,
                 session_name,    # type: AnyStr
                 session_dir,     # type: AnyStr
                 model,           # type: thelper.typedefs.ModelType
                 task,            # type: thelper.tasks.Task
                 loaders,         # type: thelper.typedefs.MultiLoaderType
                 config,          # type: thelper.typedefs.ConfigDict
                 ckptdata=None    # type: typ.Optional[thelper.typedefs.CheckpointContentType]
                 ):
        """Receives session parameters, parses image/label keys from task object, and sets up metrics."""
        super().__init__(session_name, session_dir, model, task, loaders, config, ckptdata=ckptdata)
        assert isinstance(self.task, thelper.tasks.Classification), "expected task to be classification"
        self.warned_no_shuffling_augments = False

    def _to_tensor(self, sample):
        """Fetches and returns tensors of input images and class labels from a batched sample dictionary."""
        assert isinstance(sample, dict), "trainer expects samples to come in dicts for key-based usage"
        assert self.task.input_key in sample, f"could not find input key '{self.task.input_key}' in sample dict"
        input_val, label_idx = sample[self.task.input_key], None
        if isinstance(input_val, list):
            if self.task.gt_key in sample and sample[self.task.gt_key] is not None:
                label = sample[self.task.gt_key]
                assert isinstance(label, list) and len(label) == len(input_val), \
                    "label should also be a list of the same length as input"
                label_idx = [None] * len(input_val)
                for idx in range(len(input_val)):
                    input_val[idx], label_idx[idx] = self._to_tensor({self.task.input_key: input_val[idx],
                                                                      self.task.gt_key: label[idx]})
            else:
                for idx in range(len(input_val)):
                    input_val[idx] = torch.FloatTensor(input_val[idx])
        else:
            input_val = torch.FloatTensor(input_val)
            if self.task.gt_key in sample and sample[self.task.gt_key] is not None:
                label = sample[self.task.gt_key]
                if isinstance(label, torch.Tensor) and label.numel() == input_val.shape[0] \
                        and label.dtype == torch.int64:
                    label_idx = label  # shortcut with less checks (dataset is already using tensor'd indices)
                else:
                    label_idx = label_idx or list()
                    for class_name in label:
                        assert isinstance(class_name, (int, torch.Tensor, str)), \
                            "expected label to be a name (string) or index (int)"
                        if isinstance(class_name, (int, torch.Tensor)):
                            if isinstance(class_name, torch.Tensor):
                                assert torch.numel(class_name) == 1, "unexpected scalar label, got vector"
                                class_name = class_name.item()
                            # dataset must already be using indices, we will forgive this...
                            assert 0 <= class_name < len(self.task.class_names), \
                                "class name given as out-of-range index (%d) for class list" % class_name
                            label_idx.append(class_name)
                        else:
                            assert class_name in self.task.class_names, \
                                "got unexpected label '%s' for a sample (unknown class)" % class_name
                            label_idx.append(self.task.class_indices[class_name])
                    label_idx = torch.LongTensor(label_idx)
        return input_val, label_idx

    def train_epoch(self, model, epoch, dev, loss, optimizer, loader, metrics):
        """Trains the model for a single epoch using the provided objects.

        Args:
            model: the model to train that is already uploaded to the target device(s).
            epoch: the epoch index we are training for (0-based).
            dev: the target device that tensors should be uploaded to.
            loss: the loss function used to evaluate model fidelity.
            optimizer: the optimizer used for back propagation.
            loader: the data loader used to get transformed training samples.
            metrics: the dictionary of metrics/consumers to update every iteration.
        """
        assert loss is not None, "missing loss function"
        assert optimizer is not None, "missing optimizer"
        assert loader, "no available data to load"
        assert isinstance(metrics, dict), "expect metrics as dict object"
        epoch_loss = 0
        epoch_size = len(loader)
        self.logger.debug("fetching data loader samples...")
        for idx, sample in enumerate(loader):
            input_val, label = self._to_tensor(sample)
            assert label is not None, "groundtruth required when training a model"
            optimizer.zero_grad()
            if isinstance(input_val, list):  # training samples got augmented, we need to backprop in multiple steps
                assert input_val, "cannot train with empty post-augment sample lists"
                assert isinstance(label, list) and len(label) == len(input_val), \
                    "label should also be a list of the same length as input"
                if not self.warned_no_shuffling_augments:
                    self.logger.warning("using training augmentation without global shuffling, "
                                        "gradient steps might be affected")
                    # see the docstring of thelper.transforms.operations.Duplicator for more information
                    self.warned_no_shuffling_augments = True
                iter_loss = None
                iter_pred = None
                augs_count = len(input_val)
                for input_idx in range(augs_count):
                    aug_pred = model(self._move_tensor(input_val[input_idx], dev))
                    aug_loss = loss(aug_pred, self._move_tensor(label[input_idx], dev))
                    aug_loss.backward()  # test backprop all at once? might not fit in memory...
                    if iter_pred is None:
                        iter_loss = aug_loss.clone().detach()
                        iter_pred = aug_pred.clone().detach()
                    else:
                        iter_loss += aug_loss.detach()
                        iter_pred = torch.cat((aug_pred.detach(), iter_pred), dim=0)
                iter_loss /= augs_count
                label = torch.cat(label, dim=0)
            else:
                iter_pred = model(self._move_tensor(input_val, dev))
                iter_loss = loss(iter_pred, self._move_tensor(label, dev))
                iter_loss.backward()
            optimizer.step()
            iter_pred_cpu = self._move_tensor(iter_pred, dev="cpu", detach=True)
            label_cpu = self._move_tensor(label, dev="cpu", detach=True)
            iter_loss = iter_loss.item()
            for metric in metrics.values():
                metric.update(task=self.task, input=input_val, pred=iter_pred_cpu,
                              target=label_cpu, sample=sample, loss=iter_loss, iter_idx=idx,
                              max_iters=epoch_size, epoch_idx=epoch, max_epochs=self.epochs)
            epoch_loss += iter_loss
        epoch_loss /= epoch_size
        return epoch_loss

    def eval_epoch(self, model, epoch, dev, loader, metrics):
        """Evaluates the model using the provided objects.

        Args:
            model: the model to evaluate that is already uploaded to the target device(s).
            epoch: the epoch index we are training for (0-based).
            dev: the target device that tensors should be uploaded to.
            loader: the data loader used to get transformed valid/test samples.
            metrics: the dictionary of metrics/consumers to update every iteration.
        """
        assert loader, "no available data to load"
        assert isinstance(metrics, dict), "expect metrics as dict object"
        with torch.no_grad():
            epoch_size = len(loader)
            self.logger.debug("fetching data loader samples...")
            for idx, sample in enumerate(loader):
                if idx < self.skip_eval_iter:
                    continue  # skip until previous iter count (if set externally; no effect otherwise)
                input_val, label = self._to_tensor(sample)
                if isinstance(input_val, list):  # evaluation samples got augmented, we need to get the mean prediction
                    assert input_val, "cannot eval with empty post-augment sample lists"
                    assert isinstance(label, list) and len(label) == len(input_val), \
                        "label should also be a list of the same length as input"
                    # this might be costly for nothing, we could remove the check and assume user is not dumb
                    assert not any([not torch.eq(l, label[0]).all() for l in label]), \
                        "all labels should be identical! (why do eval-time augment otherwise?)"
                    label = label[0]  # since all identical, just pick the first one and pretend its the only one
                    preds = None
                    for input_idx in range(len(input_val)):
                        pred = model(self._move_tensor(input_val[input_idx], dev))
                        if preds is None:
                            preds = torch.unsqueeze(pred.clone(), 0)
                        else:
                            preds = torch.cat((preds, torch.unsqueeze(pred, 0)), 0)
                    pred = torch.mean(preds, dim=0)
                else:
                    pred = model(self._move_tensor(input_val, dev))
                pred_cpu = self._move_tensor(pred, dev="cpu", detach=True)
                label_cpu = self._move_tensor(label, dev="cpu", detach=True)
                for metric in metrics.values():
                    metric.update(task=self.task, input=input_val, pred=pred_cpu,
                                  target=label_cpu, sample=sample, loss=None, iter_idx=idx,
                                  max_iters=epoch_size, epoch_idx=epoch, max_epochs=self.epochs)



import numpy as np

import thelper
import thelper.transforms


def fake_norm_op(sample):
    return ["norm", *sample]


def fake_norm_inv_op(sample):
    return ["norm_inv", *sample]


def fake_norm_repr_fn():
    return "Norm"


def fake_resize_op(sample):
    return ["resize", *sample]


def fake_resize_inv_op(sample):
    return ["resize_inv", *sample]


def fake_resize_repr_fn():
    return "Resize"


def test_default_compose(mocker):
    fake_norm = mocker.patch.object(thelper.transforms.operations.NormalizeMinMax, "__call__")
    fake_norm.side_effect = fake_norm_op
    fake_norm_inv = mocker.patch.object(thelper.transforms.operations.NormalizeMinMax, "invert")
    fake_norm_inv.side_effect = fake_norm_inv_op
    fake_norm_repr = mocker.patch.object(thelper.transforms.operations.NormalizeMinMax, "__repr__")
    fake_norm_repr.side_effect = fake_norm_repr_fn
    fake_resize = mocker.patch.object(thelper.transforms.operations.Resize, "__call__")
    fake_resize.side_effect = fake_resize_op
    fake_resize_inv = mocker.patch.object(thelper.transforms.operations.Resize, "invert")
    fake_resize_inv.side_effect = fake_resize_inv_op
    fake_resize_repr = mocker.patch.object(thelper.transforms.operations.Resize, "__repr__")
    fake_resize_repr.side_effect = fake_resize_repr_fn
    transforms = [
        {
            "operation": "thelper.transforms.operations.Resize",
            "params": {"dsize": (256, 256)}
        },
        {
            "operation": "thelper.transforms.operations.NormalizeMinMax",
            "params": {"min": 0, "max": 255}
        }
    ]
    composer = thelper.transforms.Compose(transforms=transforms)
    seed_call_count, epoch_call_count = 0, 0

    def fake_set_seed(seed):
        assert seed == 13
        nonlocal seed_call_count
        seed_call_count += 1

    def fake_set_epoch(epoch):
        assert epoch == 15
        nonlocal epoch_call_count
        epoch_call_count += 1

    setattr(composer[1], "set_seed", fake_set_seed)
    setattr(composer[1], "set_epoch", fake_set_epoch)
    sample_orig = ["orig"]
    sample_transf = composer(sample_orig)
    assert sample_transf == ["norm", "resize", "orig"]
    assert fake_norm.call_count == 1 and fake_resize.call_count == 1
    assert fake_norm_inv.call_count == 0 and fake_resize_inv.call_count == 0
    sample_transf_inv = composer.invert(sample_transf)
    assert sample_transf_inv == ["resize_inv", "norm_inv", "norm", "resize", "orig"]
    assert fake_norm.call_count == 1 and fake_resize.call_count == 1
    assert fake_norm_inv.call_count == 1 and fake_resize_inv.call_count == 1
    composer.set_seed(13)
    composer.set_epoch(15)
    assert seed_call_count == 1 and epoch_call_count == 1
    composer_repr = repr(composer)
    assert all([v in composer_repr for v in ["Compose", "Resize", "Norm"]])
    assert fake_norm_repr.call_count == 1 and fake_resize_repr.call_count == 1


def test_custom_compose(mocker):
    composers = thelper.transforms.CustomStepCompose(milestones={
        0: thelper.transforms.Resize(dsize=(16, 16)),
        2: thelper.transforms.Resize(dsize=(32, 32)),
        4: thelper.transforms.Resize(dsize=(64, 64)),
        8: thelper.transforms.Resize(dsize=(112, 112)),
        12: thelper.transforms.Resize(dsize=(160, 160)),
        15: thelper.transforms.Resize(dsize=(196, 196)),
        18: thelper.transforms.Resize(dsize=(224, 224)),
    }), thelper.transforms.CustomStepCompose(milestones={
        "2": [{"operation": "thelper.transforms.Resize", "params": {"dsize": (32, 32)}}],
        "4": [{"operation": "thelper.transforms.Resize", "params": {"dsize": (64, 64)}}],
        "8": [{"operation": "thelper.transforms.Resize", "params": {"dsize": (112, 112)}}],
        "12": [{"operation": "thelper.transforms.Resize", "params": {"dsize": (160, 160)}}],
        "15": [{"operation": "thelper.transforms.Resize", "params": {"dsize": (196, 196)}}],
        "18": [{"operation": "thelper.transforms.Resize", "params": {"dsize": (224, 224)}}]
    })

    def fake_set_seed(seed):
        assert seed == 13

    def fake_set_epoch(epoch):
        assert epoch >= 0

    def fake_resize_inv2(sample):
        assert sample == {}
        return "invert"

    fake_resize_inv = mocker.patch.object(thelper.transforms.operations.Resize, "invert")
    fake_resize_inv.side_effect = fake_resize_inv2
    for composer in reversed(composers):
        setattr(composer[1], "set_seed", fake_set_seed)
        setattr(composer[1], "set_epoch", fake_set_epoch)
        fake_resize_inv.reset_mock()
        for epoch in range(100):
            composer.set_seed(13)
            composer.step(epoch)
            img = np.zeros((np.random.randint(1, 1000), np.random.randint(1, 1000), 1), dtype=np.uint8)
            img_t = composer(img)
            closest_k = None
            for k, v in composer.milestones.items():
                if int(k) <= epoch:
                    closest_k = k
                else:
                    break
            if closest_k is None:
                assert img_t.shape == img.shape
            else:
                op = composer.milestones[closest_k]
                if isinstance(op, thelper.transforms.Resize):
                    assert img_t.shape[:2] == op.dsize
                else:
                    assert img_t.shape[:2] == op[0]["params"]["dsize"]
            if epoch >= 18:
                assert img_t.shape == (224, 224, 1)
        test = composer.invert({})
        assert fake_resize_inv.call_count == 1
        assert test == "invert"



"""Framework concepts module for high-level interface documenting.

The decorators and type checkers defined in this module help highlight the
purpose of classes and functions with respect to high-level ML tasks.
"""
import functools
import inspect
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, AnyStr, Callable, Optional, Union  # noqa: F401
    from types import FunctionType  # noqa: F401

SUPPORT_PREFIX = "supports_"
"""Prefix that is applied before any 'concept' decorator."""


def apply_support(func_or_cls=None, concept=None):
    # type: (Optional[Union[FunctionType, Callable]], Optional[AnyStr]) -> Callable
    """
    Utility decorator that allows marking a function or a class as *supporting* a certain ``concept``.

    Notes:
        ``concept`` support by the function or class is marked only as documentation reference, no strict
        validation is accomplished to ensure that further underlying requirements are met for the *concept*.

    .. seealso::
        | :func:`thelper.concepts.classification`
        | :func:`thelper.concepts.detection`
        | :func:`thelper.concepts.segmentation`
        | :func:`thelper.concepts.regression`
    """

    # actual function that applies the concept to the decorated 'thing'
    def apply_concept(_thing, _concept, *_args, **_kwargs):
        if isinstance(_concept, str) and len(_concept):
            _concept = f"{SUPPORT_PREFIX}{_concept}" if not _concept.startswith(SUPPORT_PREFIX) else _concept
            setattr(_thing, _concept, True)
        return _thing

    # wrapper generator for class types or class instances
    class ApplyDecorator(object):
        def __new__(cls, wrapped=None, _concept=None, *args, **kwargs):
            cls.__wrapped__ = wrapped
            return apply_concept(wrapped, _concept, *args, **kwargs)

    # wrapper generator for functions or class methods
    def apply_decorator(f_or_c):
        if inspect.isclass(f_or_c):
            return ApplyDecorator(f_or_c, concept)

        @functools.wraps(f_or_c)  # lift wrapped object definitions, so that it still looks like the original
        def decorate(*args, **kwargs):
            return apply_concept(f_or_c, concept, *args, **kwargs)
        return decorate

    if inspect.isclass(func_or_cls):
        # the '_concept' input passed down here ensures that a class decorated with a predefined
        # shortcut decorator function called without parenthesis will still receive the wanted 'concept'
        # ex:
        #   @support_classification     # <== no () here
        #   def ClassifObj(): pass
        return ApplyDecorator(func_or_cls, _concept=concept)
    elif func_or_cls:
        # this is in case the decorator is applied to a function instead of a class
        return apply_decorator(func_or_cls)
    return apply_decorator


def supports(thing, concept):
    # type: (Any, AnyStr) -> bool
    """Utility method to evaluate if ``thing`` *supports* a given ``concept`` as defined by decorators.

    Arguments:
        thing: any type, function, method, class or object instance to evaluate if it is marked by the concept
        concept: concept to check

    .. seealso::
        | :func:`thelper.concepts.classification`
        | :func:`thelper.concepts.detection`
        | :func:`thelper.concepts.segmentation`
        | :func:`thelper.concepts.regression`
    """
    return getattr(thing, f"{SUPPORT_PREFIX}{concept}", False) if isinstance(concept, str) else False


def classification(func_or_cls=None):
    # type: (Optional[Union[FunctionType, Callable]]) -> Callable
    """Decorator that allows marking a function or class as *supporting* the image classification task.

    Example::

        @thelper.concepts.classification
        class ClassifObject():
            pass

        c = ClassifObject()
        c.supports_classification
        > True

        thelper.concepts.supports(c, "classification")
        > True
    """
    return apply_support(func_or_cls, "classification")


def detection(func_or_cls=None):
    # type: (Optional[Union[FunctionType, Callable]]) -> Callable
    """Decorator that allows marking a function or class as *supporting* the object detection task.

    Example::

        @thelper.concepts.detection
        class DetectObject():
            pass

        d = DetectObject()
        d.supports_detection
        > True

        thelper.concepts.supports(d, "detection")
        > True
    """
    return apply_support(func_or_cls, "detection")


def segmentation(func_or_cls=None):
    # type: (Optional[Union[FunctionType, Callable]]) -> Callable
    """Decorator that allows marking a function or class as *supporting* the image segmentation task.

    Example::

        @thelper.concepts.segmentation
        class SegmentObject():
            pass

        s = SegmentObject()
        s.supports_segmentation
        > True

        thelper.concepts.supports(s, "segmentation")
        > True
    """
    return apply_support(func_or_cls, "segmentation")


def regression(func_or_cls=None):
    # type: (Optional[Union[FunctionType, Callable]]) -> Callable
    """Decorator that allows marking a function or class as *supporting* the generic regression task.

    Example::

        @thelper.concepts.regression
        class RegrObject():
            pass

        r = RegrObject()
        r.supports_regression
        > True

        thelper.concepts.supports(r, "regression")
        > True
    """
    return apply_support(func_or_cls, "regression")



import numpy as np
import torch

import thelper


def test_classif_logger():
    # classification results are expected in 1D format; lets build some dummy data...
    batch_size = 16
    iter_count = 32
    input_shape = (3, 32, 32)
    class_count = 10
    class_names = [str(i) for i in range(class_count)]
    task = thelper.tasks.Classification(class_names, "input", "gt", ["idx"])
    consumer_config = {"consumer": {
        "type": "thelper.train.utils.ClassifLogger",
        "params": {"top_k": 3, "report_count": 10,
                   "class_names": class_names, "log_keys": ["idx"]}
    }}
    consumers = thelper.train.create_consumers(consumer_config)
    consumer = consumers["consumer"]
    assert isinstance(consumer, thelper.train.utils.ClassifLogger)
    assert consumer.top_k == 3
    assert consumer.report_count == 10
    assert consumer.class_names == class_names
    assert consumer.report() is None
    assert repr(consumer)
    inputs, targets, preds = [], [], []
    tot_idx = 0
    for iter_idx in range(iter_count):
        # set batch size to one for 'lingering' sample in last minibatch
        curr_batch_size = batch_size if iter_idx < iter_count - 1 else 1
        inputs.append(torch.randn((curr_batch_size, *input_shape)))
        targets.append(torch.randint(low=0, high=class_count, size=(curr_batch_size, )))
        preds.append(torch.rand((curr_batch_size, class_count)))
        consumer.update(task, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                        {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                        None, iter_idx, iter_count, 0, 1)
        tot_idx += curr_batch_size
    report = consumer.report()
    assert report is not None and isinstance(report, str)
    assert len(report.split("\n")) == 11  # 10 lines + header
    assert "target_name,target_score,pred_1_name,pred_1_score,pred_2_name," \
           "pred_2_score,pred_3_name,pred_3_score,idx" == report.split("\n")[0]
    consumer.reset()
    assert consumer.report() is None
    consumer.class_names = None
    tot_idx = 0
    for iter_idx in range(iter_count):
        consumer.update(task, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                        {"idx": [tot_idx + idx for idx in range(targets[iter_idx].shape[0])]},
                        None, iter_idx, iter_count, 0, 1)
        tot_idx += targets[iter_idx].shape[0]
    assert consumer.report() == report


def test_classif_report():
    # classification results are expected in 1D format; lets build some dummy data...
    batch_size = 16
    iter_count = 32
    input_shape = (3, 32, 32)
    class_count = 10
    class_names = [str(i) for i in range(class_count)]
    task = thelper.tasks.Classification(class_names, "input", "gt", ["idx"])
    consumer_config = {"consumer": {
        "type": "thelper.train.utils.ClassifReport",
        "params": {"class_names": class_names}
    }}
    consumers = thelper.train.create_consumers(consumer_config)
    consumer = consumers["consumer"]
    assert isinstance(consumer, thelper.train.utils.ClassifReport)
    assert consumer.class_names == class_names
    assert consumer.report() is None
    assert repr(consumer)
    inputs, targets, preds = [], [], []
    tot_idx = 0
    for iter_idx in range(iter_count):
        # set batch size to one for 'lingering' sample in last minibatch
        curr_batch_size = batch_size if iter_idx < iter_count - 1 else 1
        inputs.append(torch.randn((curr_batch_size, *input_shape)))
        targets.append(torch.randint(low=0, high=class_count, size=(curr_batch_size, )))
        preds.append(torch.rand((curr_batch_size, class_count)))
        consumer.update(task, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                        {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                        None, iter_idx, iter_count, 0, 1)
        tot_idx += curr_batch_size
    report = consumer.report()
    assert report is not None and isinstance(report, str)
    assert report.endswith(f"{tot_idx}\n")  # should be total number of samples in last cell
    consumer.reset()
    assert consumer.report() is None
    consumer.class_names = None
    tot_idx = 0
    for iter_idx in range(iter_count):
        consumer.update(task, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                        {"idx": [tot_idx + idx for idx in range(targets[iter_idx].shape[0])]},
                        None, iter_idx, iter_count, 0, 1)
        tot_idx += targets[iter_idx].shape[0]
    assert consumer.report() == report


def test_confmat():
    # classification results are expected in 1D format; lets build some dummy data...
    batch_size = 16
    iter_count = 32
    input_shape = (3, 32, 32)
    class_count = 10
    class_names = [str(i) for i in range(class_count)]
    task = thelper.tasks.Classification(class_names, "input", "gt", ["idx"])
    consumer_config = {"consumer": {
        "type": "thelper.train.utils.ConfusionMatrix",
        "params": {"class_names": class_names}
    }}
    consumers = thelper.train.create_consumers(consumer_config)
    consumer = consumers["consumer"]
    assert isinstance(consumer, thelper.train.utils.ConfusionMatrix)
    assert consumer.class_names == class_names
    assert consumer.report() is None
    assert repr(consumer)
    inputs, targets, preds = [], [], []
    tot_idx = 0
    for iter_idx in range(iter_count):
        # set batch size to one for 'lingering' sample in last minibatch
        curr_batch_size = batch_size if iter_idx < iter_count - 1 else 1
        inputs.append(torch.randn((curr_batch_size, *input_shape)))
        targets.append(torch.randint(low=0, high=class_count, size=(curr_batch_size, )))
        preds.append(torch.rand((curr_batch_size, class_count)))
        consumer.update(task, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                        {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                        None, iter_idx, iter_count, 0, 1)
        tot_idx += curr_batch_size
    report = consumer.report()
    assert report is not None and isinstance(report, str)
    assert report.endswith(f"{tot_idx}\n")  # should be total number of samples in last cell
    render = consumer.render()
    assert render is None or isinstance(render, np.ndarray)
    consumer.reset()
    assert consumer.report() is None
    consumer.class_names = None
    tot_idx = 0
    for iter_idx in range(iter_count):
        consumer.update(task, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                        {"idx": [tot_idx + idx for idx in range(targets[iter_idx].shape[0])]},
                        None, iter_idx, iter_count, 0, 1)
        tot_idx += targets[iter_idx].shape[0]
    assert consumer.report() == report



"""Regression task interface module.

This module contains a class that defines the objectives of models/trainers for regression tasks.
"""
import logging
from typing import Optional  # noqa: F401

import numpy as np

import thelper.concepts
import thelper.utils
from thelper.tasks.utils import Task

logger = logging.getLogger(__name__)


@thelper.concepts.regression
class Regression(Task):
    """Interface for n-dimension regression tasks.

    This specialization requests that when given an input tensor, the trained model should
    provide an n-dimensional target prediction. This is a fairly generic task that (unlike
    image classification and semantic segmentation) is not linked to a pre-existing set of
    possible solutions. The task interface is used to carry useful metadata for this task,
    e.g. input/output shapes, types, and min/max values for rounding/saturation.

    Attributes:
        input_shape: a numpy-compatible shape to expect model inputs to be in.
        target_shape: a numpy-compatible shape to expect the predictions to be in.
        target_type: a numpy-compatible type to cast the predictions to (if needed).
        target_min: an n-dim tensor containing minimum target values (if applicable).
        target_max: an n-dim tensor containing maximum target values (if applicable).
        input_key: the key used to fetch input tensors from a sample dictionary.
        target_key: the key used to fetch target (groundtruth) values from a sample dictionary.
        meta_keys: the list of extra keys provided by the data parser inside each sample.

    .. seealso::
        | :class:`thelper.tasks.utils.Task`
        | :class:`thelper.train.regr.RegressionTrainer`
        | :class:`thelper.tasks.regr.SuperResolution`
        | :class:`thelper.tasks.detect.Detection`
    """

    def __init__(self, input_key, target_key, meta_keys=None, input_shape=None,
                 target_shape=None, target_type=None, target_min=None, target_max=None):
        """Receives and stores the keys produced by the dataset parser(s)."""
        super(Regression, self).__init__(input_key, target_key, meta_keys)
        self.input_shape = input_shape
        self.target_shape = target_shape
        self.target_type = target_type
        self.target_min = target_min
        self.target_max = target_max
        if self.target_type is not None:
            assert not isinstance(self.target_min, np.ndarray) or self.target_min.dtype == self.target_type, \
                "invalid target min dtype"
            assert not isinstance(self.target_max, np.ndarray) or self.target_max.dtype == self.target_type, \
                "invalid target max dtype"
        if self.target_shape is not None:
            assert not isinstance(self.target_min, np.ndarray) or self.target_min.shape == self.target_shape, \
                "invalid target min shape"
            assert not isinstance(self.target_max, np.ndarray) or self.target_max.shape == self.target_shape, \
                "invalid target max shape"
        if isinstance(self.target_min, np.ndarray) and isinstance(self.target_max, np.ndarray):
            assert self.target_min.shape == self.target_max.shape, "target min/max shape mismatch"

    @property
    def input_shape(self):
        """Returns the shape of the input tensors to be processed by the model."""
        return self._input_shape

    @input_shape.setter
    def input_shape(self, input_shape):
        """Sets the shape of the input tensors to be processed by the model."""
        if input_shape is not None:
            if isinstance(input_shape, list):
                input_shape = tuple(input_shape)
            assert isinstance(input_shape, tuple) and all([isinstance(v, int) for v in input_shape]), \
                "unexpected input shape type (should be tuple of integers)"
        self._input_shape = input_shape

    @property
    def target_shape(self):
        """Returns the shape of the output tensors to be generated by the model."""
        return self._target_shape

    @target_shape.setter
    def target_shape(self, target_shape):
        """Sets the shape of the output tensors to be generated by the model."""
        if target_shape is not None:
            if isinstance(target_shape, list):
                target_shape = tuple(self.target_shape)
            assert isinstance(target_shape, tuple) and all([isinstance(v, int) for v in target_shape]), \
                "unexpected target shape type (should be tuple of integers)"
        self._target_shape = target_shape

    @property
    def target_type(self):
        """Returns the type of the output tensors to be generated by the model."""
        return self._target_type

    @target_type.setter
    def target_type(self, target_type):
        """Sets the type of the output tensors to be generated by the model."""
        if target_type is not None:
            if isinstance(target_type, str):
                import thelper.utils
                target_type = thelper.utils.import_class(target_type)
            assert issubclass(target_type, np.generic), "target type should be a numpy-compatible type"
        self._target_type = target_type

    @property
    def target_min(self):
        """Returns the minimum target value(s) to be generated by the model."""
        return self._target_min

    @target_min.setter
    def target_min(self, target_min):
        """Sets the minimum target value(s) to be generated by the model."""
        if target_min is not None:
            if isinstance(target_min, (list, tuple)):
                target_min = np.asarray(target_min)
            assert isinstance(target_min, np.ndarray), "target_min should be passed as list/tuple/ndarray"
        self._target_min = target_min

    @property
    def target_max(self):
        """Returns the maximum target value(s) to be generated by the model."""
        return self._target_max

    @target_max.setter
    def target_max(self, target_max):
        """Sets the maximum target value(s) to be generated by the model."""
        if target_max is not None:
            if isinstance(target_max, (list, tuple)):
                target_max = np.asarray(target_max)
            assert isinstance(target_max, np.ndarray), "target_max should be passed as list/tuple/ndarray"
        self._target_max = target_max

    def check_compat(self, task, exact=False):
        # type: (Regression, Optional[bool]) -> bool
        """Returns whether the current task is compatible with the provided one or not.

        This is useful for sanity-checking, and to see if the inputs/outputs of two models
        are compatible. If ``exact = True``, all fields will be checked for exact (perfect)
        compatibility (in this case, matching meta keys).
        """
        if isinstance(task, Regression):
            # if both tasks are related to regression: all non-None keys and specs must match
            return (self.input_key == task.input_key and
                    (self.gt_key is None or task.gt_key is None or self.gt_key == task.gt_key) and
                    (self.input_shape is None or task.input_shape is None or self.input_shape == task.input_shape) and
                    (self.target_shape is None or task.target_shape is None or self.target_shape == task.target_shape) and
                    (self.target_type is None or task.target_type is None or self.target_type == task.target_type) and
                    (self.target_min is None or task.target_min is None or self.target_min == task.target_min) and
                    (self.target_max is None or task.target_max is None or self.target_max == task.target_max) and
                    (not exact or (set(self.meta_keys) == set(task.meta_keys) and
                                   self.gt_key == task.gt_key and
                                   self.input_shape == task.input_shape and
                                   self.target_shape == task.target_shape and
                                   self.target_type == task.target_type and
                                   self.target_min == task.target_min and
                                   self.target_max == task.target_max)))
        elif type(task) == Task:
            # if 'task' simply has no gt, compatibility rests on input key only
            return not exact and self.input_key == task.input_key and task.gt_key is None
        return False

    def get_compat(self, task):
        """Returns a task instance compatible with the current task and the given one."""
        # currently not checking for input/target param intersections between similar regression tasks
        assert self.check_compat(task), f"cannot create compatible task between:\n\t{str(self)}\n\t{str(task)}"
        meta_keys = list(set(self.meta_keys + task.meta_keys))
        return Regression(input_key=self.input_key, target_key=self.gt_key, meta_keys=meta_keys,
                          input_shape=self.input_shape if self.input_shape is not None else task.input_shape,
                          target_shape=self.target_shape if self.target_shape is not None else task.target_shape,
                          target_type=self.target_type if self.target_type is not None else task.target_type,
                          target_min=self.target_min if self.target_min is not None else task.target_min,
                          target_max=self.target_max if self.target_max is not None else task.target_max)

    def __repr__(self):
        """Creates a print-friendly representation of a segmentation task."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(input_key={repr(self.input_key)}, target_key={repr(self.gt_key)}, " + \
            f"meta_keys={repr(self.meta_keys)}, input_shape={repr(self.input_shape)}, " + \
            f"target_shape={repr(self.target_shape)}, target_type={repr(self.target_type)}, " + \
            f"target_min={repr(self.target_min)}, target_max={repr(self.target_max)})"


@thelper.concepts.regression
class SuperResolution(Regression):
    """Interface for super-resolution tasks.

    This specialization requests that when given an input tensor, the trained model should
    provide an identically-shape target prediction that essentially contains more (or more
    adequate) high-frequency spatial components.

    This specialized regression interface is currently used to help display functions.

    Attributes:
        input_shape: a numpy-compatible shape to expect model inputs/outputs to be in.
        target_type: a numpy-compatible type to cast the predictions to (if needed).
        target_min: an n-dim tensor containing minimum target values (if applicable).
        target_max: an n-dim tensor containing maximum target values (if applicable).
        input_key: the key used to fetch input tensors from a sample dictionary.
        target_key: the key used to fetch target (groundtruth) values from a sample dictionary.
        meta_keys: the list of extra keys provided by the data parser inside each sample.

    .. seealso::
        | :class:`thelper.tasks.utils.Task`
        | :class:`thelper.tasks.regr.Regression`
        | :class:`thelper.train.regr.RegressionTrainer`
    """

    def __init__(self, input_key, target_key, meta_keys=None, input_shape=None, target_type=None,
                 target_min=None, target_max=None):
        """Receives and stores the keys produced by the dataset parser(s)."""
        super(SuperResolution, self).__init__(input_key, target_key, meta_keys,
                                              input_shape=input_shape, target_shape=input_shape,
                                              target_type=target_type, target_min=target_min,
                                              target_max=target_max)



"""Object detection trainer/evaluator implementation module."""
import collections
import logging
from typing import AnyStr  # noqa: F401

import numpy as np
import torch
import torch.optim
import torchvision

import thelper.concepts
import thelper.typedefs as typ  # noqa: F401
import thelper.utils
from thelper.train.base import Trainer

logger = logging.getLogger(__name__)


@thelper.concepts.detection
class ObjDetectTrainer(Trainer):
    """Trainer interface specialized for object detection.

    This class implements the abstract functions of :class:`thelper.train.base.Trainer` required to train/evaluate
    a model for object detection (i.e. 2D bounding box regression). It also provides a utility function for fetching
    i/o packets (input images, bounding boxes) from a sample, and that converts those into tensors for forwarding
    and loss estimation.

    .. seealso::
        | :class:`thelper.train.base.Trainer`
    """

    def __init__(self,
                 session_name,    # type: AnyStr
                 session_dir,     # type: AnyStr
                 model,           # type: thelper.typedefs.ModelType
                 task,            # type: thelper.tasks.Task
                 loaders,         # type: thelper.typedefs.MultiLoaderType
                 config,          # type: thelper.typedefs.ConfigDict
                 ckptdata=None    # type: typ.Optional[thelper.typedefs.CheckpointContentType]
                 ):
        """Receives session parameters, parses tensor/target keys from task object, and sets up metrics."""
        super().__init__(session_name, session_dir, model, task, loaders, config, ckptdata=ckptdata)
        assert isinstance(self.task, thelper.tasks.Detection), "expected task to be object detection"

    def _to_tensor(self, sample):
        """Fetches and returns tensors of inputs and targets from a batched sample dictionary."""
        assert isinstance(sample, dict), "trainer expects samples to come in dicts for key-based usage"
        assert self.task.input_key in sample, f"could not find input key '{self.task.input_key}' in sample dict"
        input_val = sample[self.task.input_key]
        if isinstance(input_val, np.ndarray):
            input_val = torch.from_numpy(input_val)
        assert isinstance(input_val, torch.Tensor), "unexpected input type; should be torch.Tensor"
        if self.task.input_shape is not None:
            assert input_val.dim() == len(self.task.input_shape) + 1, \
                "expected input as Nx[shape] where N = batch size"
            assert self.task.input_shape == input_val.shape[1:], \
                f"invalid input shape; got '{input_val.shape[1:]}', expected '{self.task.input_shape}'"
        assert input_val.dim() == 4, "input image stack should be 4-dim to be decomposed into list of images"
        # unpack input images into list (as required by torchvision preproc)
        input_val = [input_val[i] for i in range(input_val.shape[0])]
        bboxes = None
        if self.task.gt_key in sample:
            bboxes = sample[self.task.gt_key]
            assert isinstance(bboxes, list) and all([isinstance(bset, list) for bset in bboxes]), \
                "bboxes should be provided as a list of lists (dims = batch x bboxes-per-image)"
            assert all([all([isinstance(box, thelper.data.BoundingBox) for box in bset]) for bset in bboxes]), \
                "bboxes should be provided as a thelper.data.BoundingBox-compat object"
            assert all([len(np.unique([b.image_id for b in bset if b.image_id is not None])) <= 1 for bset in bboxes]), \
                "some bboxes tied to a single image have different reference ids"
            # here, we follow the format used in torchvision (>=0.3) for forwarding targets to detection models
            # (see https://pytorch.org/tutorials/intermediate/torchvision_tutorial.html for more info)
            bboxes = [{
                "boxes": torch.as_tensor([[*b.bbox] for b in bset], dtype=torch.float32),
                "labels": torch.as_tensor([b.class_id for b in bset], dtype=torch.int64),
                "image_id": torch.as_tensor([b.image_id if b.image_id is not None else -1 for b in bset]),
                "area": torch.as_tensor([b.area for b in bset], dtype=torch.float32),
                "iscrowd": torch.as_tensor([b.iscrowd for b in bset], dtype=torch.int64),
                "refs": bset
            } for bset in bboxes]
        return input_val, bboxes

    def _from_tensor(self, bboxes, sample=None):
        """Fetches and returns a list of bbox objects from a model-specific representation."""
        # for now, we can only unpack torchvision-format bbox dictionary lists (everything else will throw)
        assert isinstance(bboxes, list), "input should be list since we do batch predictions"
        if all([isinstance(d, dict) and len(d) == 3 and
                all([k in ["boxes", "labels", "scores"] for k in d]) for d in bboxes]):
            outputs = []
            for batch_idx, d in enumerate(bboxes):
                boxes = d["boxes"].detach().cpu()
                labels = d["labels"].detach().cpu()
                scores = d["scores"].detach().cpu()
                assert boxes.shape[0] == labels.shape[0] and boxes.shape[0] == scores.shape[0], "mismatched tensor dims"
                curr_output = []
                for box_idx, box in enumerate(boxes):
                    if sample is not None and self.task.gt_key in sample and sample[self.task.gt_key][batch_idx]:
                        gt_box = sample[self.task.gt_key][batch_idx][0]  # use first gt box to get image-level props
                        out = thelper.data.BoundingBox(labels[box_idx].item(), box, confidence=scores[box_idx].item(),
                                                       image_id=gt_box.image_id, task=self.task)

                    elif sample is not None and "idx" in sample:
                        out = thelper.data.BoundingBox(labels[box_idx].item(), box, confidence=scores[box_idx].item(),
                                                       image_id=sample["idx"][batch_idx], task=self.task)
                    else:
                        out = thelper.data.BoundingBox(labels[box_idx].item(), box, confidence=scores[box_idx].item(),
                                                       task=self.task)
                    curr_output.append(out)
                outputs.append(curr_output)
            return outputs
        raise AssertionError("unrecognized packed bboxes vector format")

    def train_epoch(self, model, epoch, dev, loss, optimizer, loader, metrics):
        """Trains the model for a single epoch using the provided objects.

        Args:
            model: the model to train that is already uploaded to the target device(s).
            epoch: the epoch index we are training for (0-based).
            dev: the target device that tensors should be uploaded to.
            loss: the loss function used to evaluate model fidelity.
            optimizer: the optimizer used for back propagation.
            loader: the data loader used to get transformed training samples.
            metrics: the dictionary of metrics/consumers to update every iteration.
        """
        assert loss is None, "current implementation assumes that loss is computed inside the model"
        assert optimizer is not None, "missing optimizer"
        assert loader, "no available data to load"
        assert isinstance(metrics, dict), "expect metrics as dict object"
        epoch_loss = 0
        epoch_size = len(loader)
        self.logger.debug("fetching data loader samples...")
        for idx, sample in enumerate(loader):
            images, targets = self._to_tensor(sample)
            assert targets is not None and not any([not bset for bset in targets]), \
                "groundtruth required when training a model"
            optimizer.zero_grad()
            targets_dev = self._move_tensor(targets, dev)
            images_dev = self._move_tensor(images, dev)
            if isinstance(model, thelper.nn.utils.ExternalModule):
                model = model.model  # temporarily unwrap to simplify code below
            assert isinstance(model, torchvision.models.detection.generalized_rcnn.GeneralizedRCNN), \
                "unknown/unhandled detection model type"  # user should probably implement their own trainer
            # unfortunately, the default generalized RCNN model forward does not return predictions while training...
            # loss_dict = model(images=images_dev, targets=targets)  # we basically reimplement this call below
            original_image_sizes = [img.shape[-2:] for img in images_dev]
            images_dev, targets_dev = model.transform(images_dev, targets_dev)
            features = model.backbone(images_dev.tensors)
            if isinstance(features, torch.Tensor):
                features = collections.OrderedDict([(0, features)])
            proposals, proposal_losses = model.rpn(images_dev, features, targets_dev)
            pred, pred_losses = model.roi_heads(features, proposals, images_dev.image_sizes, targets_dev)
            pred = model.transform.postprocess(pred, images_dev.image_sizes, original_image_sizes)
            iter_loss = sum(loss for loss in {**pred_losses, **proposal_losses}.values())
            iter_loss.backward()
            optimizer.step()
            pred = self._from_tensor(pred, sample)
            target_bboxes = [target["refs"] for target in targets]
            # pack image list back into 4d tensor
            images = torch.cat(images) if len(images) > 1 else torch.unsqueeze(images[0], 0)
            iter_loss = iter_loss.item()
            if not pred:  # for some reason, preds are not provided in train mode by faster-rcnn head
                pred = [[] * images.shape[0]]  # ... create dummy list of lists for metrics
            for metric in metrics.values():
                metric.update(task=self.task, input=images, pred=pred, target=target_bboxes,
                              sample=sample, loss=iter_loss, iter_idx=idx, max_iters=epoch_size,
                              epoch_idx=epoch, max_epochs=self.epochs)
            epoch_loss += iter_loss
        epoch_loss /= epoch_size
        return epoch_loss

    def eval_epoch(self, model, epoch, dev, loader, metrics):
        """Evaluates the model using the provided objects.

        Args:
            model: the model to evaluate that is already uploaded to the target device(s).
            epoch: the epoch index we are training for (0-based).
            dev: the target device that tensors should be uploaded to.
            loader: the data loader used to get transformed valid/test samples.
            metrics: the dictionary of metrics/consumers to update every iteration.
        """
        assert loader, "no available data to load"
        assert isinstance(metrics, dict), "expect metrics as dict object"
        with torch.no_grad():
            epoch_size = len(loader)
            self.logger.debug("fetching data loader samples...")
            for idx, sample in enumerate(loader):
                if idx < self.skip_eval_iter:
                    continue  # skip until previous iter count (if set externally; no effect otherwise)
                images, targets = self._to_tensor(sample)
                pred = model(self._move_tensor(images, dev))
                pred = self._from_tensor(pred, sample)
                target_bboxes = [target["refs"] for target in targets]
                # pack image list back into 4d tensor
                images = torch.cat(images) if len(images) > 1 else torch.unsqueeze(images[0], 0)
                for metric in metrics.values():
                    metric.update(task=self.task, input=images, pred=pred, target=target_bboxes,
                                  sample=sample, loss=None, iter_idx=idx, max_iters=epoch_size,
                                  epoch_idx=epoch, max_epochs=self.epochs)



import os
import shutil

import numpy as np
import pytest
import torch

import thelper
import thelper.nn.resnet

test_save_path = ".pytest_cache"
test_classif_leak_name = "test-classif-leak"

test_classif_leak_path = os.path.join(test_save_path, test_classif_leak_name)


def check_gpu_compat():
    return torch.cuda.is_available() and \
        torch.cuda.get_device_properties(torch.device("cuda")).total_memory >= 4 * 1024 * 1024 * 1024


@pytest.mark.skipif(not check_gpu_compat(), reason="test requires GPU w/ >4GB RAM")
def test_model_leak():
    device = torch.device("cuda")
    for test_idx in range(25):
        model = thelper.nn.resnet.ResNet(thelper.tasks.Classification([str(idx) for idx in range(1000)], "0", "1")).to(device)
        model.train()
        loss_fn = torch.nn.CrossEntropyLoss()
        optimizer = torch.optim.Adam(model.parameters())
        batch_size = 40
        for iter_idx in range(5):
            optimizer.zero_grad()
            pred = model(torch.randn(batch_size, 3, 224, 224).to(device))
            loss_fn(pred, torch.randint(1000, size=(batch_size,)).to(device)).backward()
            optimizer.step()
        del pred
        del optimizer
        del model


class DummyDataset(thelper.data.Dataset):
    def __init__(self, nb_samples, transforms=None, deepcopy=None):
        super().__init__(transforms=transforms, deepcopy=deepcopy)
        self.samples = np.arange(nb_samples)
        self.task = thelper.tasks.Classification([str(idx) for idx in range(1000)], "0", "1")

    def __getitem__(self, idx):  # pragma: no cover
        return {"0": torch.randn(3, 224, 224), "1": torch.randint(1000, size=())}


@pytest.fixture
def config(request):
    def fin():
        shutil.rmtree(test_classif_leak_path, ignore_errors=True)
    fin()
    request.addfinalizer(fin)
    return {
        "name": test_classif_leak_name,
        "bypass_queries": True,
        "datasets": {
            "dummy": DummyDataset(100)
        },
        "loaders": {
            "shuffle": True,
            "batch_size": 35,
            "skip_class_balancing": True,
            "train_split": {
                "dummy": 0.9
            },
            "valid_split": {
                "dummy": 0.1
            }
        },
        "model": {
            "type": "thelper.nn.resnet.ResNet"
        },
        "trainer": {
            "type": "thelper.train.ImageClassifTrainer",
            "epochs": 1,
            "monitor": "accuracy",
            "optimization": {
                "loss": {
                    "type": "torch.nn.CrossEntropyLoss"
                },
                "optimizer": {
                    "type": "torch.optim.Adam",
                    "params": {
                        "lr": 0.001
                    }
                }
            },
            "metrics": {
                "accuracy": {
                    "type": "thelper.optim.Accuracy",
                    "params": {
                        "top_k": 1
                    }
                }
            }
        }
    }


@pytest.mark.skipif(not check_gpu_compat(), reason="test requires GPU w/ >4GB RAM")
def test_train_decomposed_leak(config, mocker):
    mocker.patch("json.dump")
    mocker.patch("torch.save")
    session_name = config["name"]
    save_dir = thelper.utils.get_save_dir(test_save_path, session_name, config)
    task, train_loader, valid_loader, test_loader = thelper.data.create_loaders(config, save_dir)
    loaders = (train_loader, valid_loader, test_loader)
    outputs = []
    for idx in range(5):
        model = thelper.nn.create_model(config, task, save_dir=save_dir)
        trainer = thelper.train.create_trainer(session_name, save_dir, config, model, task, loaders)
        trainer.train()
        outputs.append(trainer.outputs)


@pytest.mark.skipif(not check_gpu_compat(), reason="test requires GPU w/ >4GB RAM")
def test_train_leak(config, mocker):
    mocker.patch("json.dump")
    mocker.patch("torch.save")
    for idx in range(5):
        train_outputs = thelper.cli.create_session(config, test_save_path)
        del train_outputs



"""
Tests support concept decorators
"""
import thelper.concepts


def test_class_decorated_with_parenthesis_and_concept():

    @thelper.concepts.apply_support(concept="test1")
    class Test1(object):
        pass

    @thelper.concepts.apply_support(concept="test2")
    class Test2(Test1):
        pass

    class Test3(Test1):
        pass

    assert thelper.concepts.supports(Test1, "test1")
    assert thelper.concepts.supports(Test2, "test1")
    assert thelper.concepts.supports(Test2, "test2")
    assert thelper.concepts.supports(Test3, "test1")
    assert not thelper.concepts.supports(Test3, "test2")

    t1 = Test1()
    t2 = Test2()
    t3 = Test3()
    assert thelper.concepts.supports(t1, "test1")
    assert thelper.concepts.supports(t2, "test1")
    assert thelper.concepts.supports(t2, "test2")
    assert thelper.concepts.supports(t3, "test1")
    assert not thelper.concepts.supports(t3, "test2")


def test_class_decorated_with_helper_and_parenthesis():
    def helper1(func=None):
        return thelper.concepts.apply_support(func, "test1")

    def helper2(func=None):
        return thelper.concepts.apply_support(func, "test2")

    @helper1()
    class Test1(object):
        pass

    @helper2()
    class Test2(Test1):
        pass

    class Test3(Test1):
        pass

    assert thelper.concepts.supports(Test1, "test1")
    assert thelper.concepts.supports(Test2, "test1")
    assert thelper.concepts.supports(Test2, "test2")
    assert thelper.concepts.supports(Test3, "test1")
    assert not thelper.concepts.supports(Test3, "test2")

    t1 = Test1()
    t2 = Test2()
    t3 = Test3()
    assert thelper.concepts.supports(t1, "test1")
    assert thelper.concepts.supports(t2, "test1")
    assert thelper.concepts.supports(t2, "test2")
    assert thelper.concepts.supports(t3, "test1")
    assert not thelper.concepts.supports(t3, "test2")


def test_class_decorated_with_helper_without_parenthesis():
    def helper1(func=None):
        return thelper.concepts.apply_support(func, "test1")

    def helper2(func=None):
        return thelper.concepts.apply_support(func, "test2")

    @helper1    # no parenthesis here
    class Test1(object):
        pass

    @helper2    # no parenthesis here
    class Test2(Test1):
        pass

    class Test3(Test1):
        pass

    assert thelper.concepts.supports(Test1, "test1")
    assert thelper.concepts.supports(Test2, "test1")
    assert thelper.concepts.supports(Test2, "test2")
    assert thelper.concepts.supports(Test3, "test1")
    assert not thelper.concepts.supports(Test3, "test2")

    t1 = Test1()
    t2 = Test2()
    t3 = Test3()
    assert thelper.concepts.supports(t1, "test1")
    assert thelper.concepts.supports(t2, "test1")
    assert thelper.concepts.supports(t2, "test2")
    assert thelper.concepts.supports(t3, "test1")
    assert not thelper.concepts.supports(t3, "test2")



"""Dataset parsers module.

This module contains dataset parser interfaces and base classes that define basic i/o
operations so that the framework can automatically interact with training data.
"""

import inspect
import logging
import os
from abc import abstractmethod

import cv2 as cv
import numpy as np
import PIL
import PIL.Image
import torch
import torch.utils.data

import thelper.tasks
import thelper.utils

logger = logging.getLogger(__name__)


class Dataset(torch.utils.data.Dataset):
    """Abstract dataset parsing interface that holds a task and a list of sample dictionaries.

    This interface helps fix a failure of PyTorch's dataset interface (``torch.utils.data.Dataset``):
    the lack of identity associated with the components of a sample. In short, a data sample loaded by a
    dataset typically contains the input data that should be forwarded to a model as well as the expected
    prediction of the model (i.e. the 'groundtruth') that will be used to compute the loss. These two
    elements are typically paired in a tuple that can then be provided to the data loader for batching.
    Problems however arise when the model has multiple inputs or outputs, when the sample needs to carry
    supplemental metadata to simplify debugging, or when transformation operations need to be applied
    only to specific elements of the sample. Here, we fix this issue by specifying that all samples must
    be provided to data loaders as dictionaries. The keys of these dictionaries explicitly define which
    value(s) should be transformed, which should be forwarded to the model, which are the expected model
    predictions, and which are only used for debugging. The keys are defined via the task object that is
    generated by the dataset or specified via the configuration file (see :class:`thelper.tasks.utils.Task`
    for more information).

    To properly use this interface, a derived class must thus implement :func:`thelper.data.parsers.Dataset.__getitem__`,
    provide proper ``self.task`` and ``self.samples`` attributes.

    Attributes:
        transforms: function or object that should be applied to all loaded samples in order to
            return the data in the requested transformed/augmented state.
        deepcopy: specifies whether this dataset interface should be deep-copied inside
            :class:`thelper.data.loaders.LoaderFactory` so that it may be shared between
            different threads. This is false by default, as we assume datasets do not contain a state
            or buffer that might cause problems in multi-threaded data loaders.
        samples: list of dictionaries containing the data that is ready to be forwarded to the
            data loader. Note that relatively costly operations (such as reading images from a disk
            or pre-transforming them) should be delayed until the :func:`thelper.data.parsers.Dataset.__getitem__`
            function is called, as they will most likely then be accomplished in a separate thread.
            Once loaded, these samples should never be modified by another part of the framework. For
            example, transformation and augmentation operations will always be applied to copies
            of these samples.
        task: object used to define what keys are used to index the loaded data into sample dictionaries.

    .. seealso::
        | :class:`thelper.data.parsers.ExternalDataset`
    """

    def __init__(self, transforms=None, deepcopy=False):
        """Dataset parser constructor.

        In order for derived datasets to be instantiated automatically by the framework from a configuration
        file, they must minimally accept a 'transforms' argument like the shown one here.

        Args:
            transforms: function or object that should be applied to all loaded samples in order to
                return the data in the requested transformed/augmented state.
            deepcopy: specifies whether this dataset interface should be deep-copied inside
                :class:`thelper.data.loaders.LoaderFactory` so that it may be shared between
                different threads. This is false by default, as we assume datasets do not contain a state
                or buffer that might cause problems in multi-threaded data loaders.
        """
        super(Dataset, self).__init__()
        self.transforms = transforms
        self.deepcopy = deepcopy  # will determine if we deepcopy in each loader
        self.samples = None  # must be set by the derived class as a list of dictionaries
        self.task = None  # must be set by the derived class as a valid task object

    def _get_derived_name(self):
        """Returns a pretty-print version of the derived class's name."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__

    def __len__(self):
        """Returns the total number of samples available from this dataset interface."""
        return len(self.samples)

    def __iter__(self):
        """Returns an iterator over the dataset's samples."""
        for idx in range(len(self.samples)):
            yield self[idx]

    @abstractmethod
    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        raise NotImplementedError

    @property
    def transforms(self):
        """Returns the transformation operations to apply to this dataset's loaded samples."""
        return self._transforms

    @transforms.setter
    def transforms(self, transforms):
        """Sets the transformation operations to apply to this dataset's loaded samples."""
        assert transforms is None or hasattr(transforms, "__call__") or \
            (isinstance(transforms, list) and all([hasattr(t, "__call__") for t in transforms])), \
            "transformations should be callable (or list of callables)"
        self._transforms = transforms

    @property
    def deepcopy(self):
        """specifies whether this dataset interface should be deep-copied inside
        :class:`thelper.data.loaders.LoaderFactory` so that it may be shared between
        different threads. This is false by default, as we assume datasets do not contain a state
        or buffer that might cause problems in multi-threaded data loaders."""
        return self._deepcopy

    @deepcopy.setter
    def deepcopy(self, deepcopy):
        assert deepcopy is None or isinstance(deepcopy, bool), "deepcopy flag should be boolean"
        self._deepcopy = deepcopy

    @property
    def task(self):
        """Returns the task object associated with this dataset interface."""
        return self._task

    @task.setter
    def task(self, task):
        """Sets the task object associated with this dataset interface."""
        assert task is None or isinstance(task, thelper.tasks.Task), "invalid task"
        self._task = task

    @property
    def samples(self):
        """Returns the list of internal samples held by this dataset interface."""
        return self._samples

    @samples.setter
    def samples(self, samples):
        assert samples is None or isinstance(samples, (list, Dataset)) or hasattr(samples, "__getitem__"), \
            "invalid samples (should be list of dicts, dataset, or it should have '__getitem__' attrib)"
        self._samples = samples

    def _getitems(self, idxs):
        """Returns a list of dictionaries corresponding to the sliced sample indices."""
        if not isinstance(idxs, slice):
            raise AssertionError("unexpected input (should be slice)")
        return [self[idx] for idx in range(*idxs.indices(len(self)))]

    def __repr__(self):
        """Returns a print-friendly representation of this dataset."""
        return self._get_derived_name() + f"(transforms={repr(self.transforms)}, deepcopy={repr(self.deepcopy)})"


class HDF5Dataset(Dataset):
    """HDF5 dataset specialization interface.

    This specialization is compatible with the HDF5 packages made by the CLI's "split" operation. The
    archives it loads contains pre-split datasets that can be reloaded without having to resplit their
    data. The archive also contains useful metadata, and a task interface.

    Attributes:
        archive: file descriptor for the opened hdf5 dataset.
        subset: hdf5 group section representing the targeted set.
        target_args: list decompression args required for each sample key.
        source: source logstamp of the hdf5 dataset.
        git_sha1: framework git tag of the hdf5 dataset.
        version: version of the framework that saved the hdf5 dataset.
        orig_config: configuration used to originally generate the hdf5 dataset.

    .. seealso::
        | :func:`thelper.cli.split_data`
        | :func:`thelper.data.utils.create_hdf5`
    """

    def __init__(self, root, subset="train", transforms=None):
        """HDF5 dataset parser constructor.

        This constructor receives the path to the HDF5 archive as well as a subset indicating which
        section of the archive to load. By default, it loads the training set.
        """
        super(HDF5Dataset, self).__init__(transforms=transforms, deepcopy=False)
        assert subset in ["train", "valid", "test"], f"unrecognized subset '{subset}'"
        import h5py
        self.archive = h5py.File(root, "r")
        self.source = self.archive.attrs["source"]
        self.git_sha1 = self.archive.attrs["git_sha1"]
        self.version = self.archive.attrs["version"]
        self.task = thelper.tasks.create_task(self.archive.attrs["task"])
        self.orig_config = eval(self.archive.attrs["config"])
        compr_config = eval(self.archive.attrs["compression"])
        if subset not in self.archive:
            raise AssertionError(f"subset '{subset}' not found in hdf5 archive")
        self.subset = self.archive[subset]
        sample_count = self.subset.attrs["count"]
        self.samples = [{}] * sample_count
        self.target_args = {}
        for key in self.task.keys:
            dset = self.subset[key]
            assert dset.len() == len(self.samples)
            dtype = dset.attrs["orig_dtype"] if "orig_dtype" in dset.attrs else None
            shape = dset.attrs["orig_shape"] if "orig_shape" in dset.attrs else None
            compr_config = thelper.utils.get_key_def(key, compr_config, default={})
            compr_type = thelper.utils.get_key_def("type", compr_config, default="none")
            compr_kwargs = thelper.utils.get_key_def(["decode_params", "decode_kwargs"], compr_config, default={})
            self.target_args[key] = {"dset": dset, "dtype": dtype, "shape": shape, "compr_type": compr_type, "compr_kwargs": compr_kwargs}

    def _unpack(self, dset, idx, dtype=None, shape=None, compr_type="none", **compr_kwargs):
        if dtype is not None:
            array = np.frombuffer(thelper.utils.decode_data(dset[idx], compr_type, **compr_kwargs), dtype=dtype)
        else:
            array = dset[idx]
        if shape is not None:
            if np.issubdtype(dtype, np.dtype(str).type) and len(shape) == 0:
                array = "".join(array)  # reassemble string if needed
            else:
                array = array.reshape(shape)
        return array

    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        if isinstance(idx, slice):
            return self._getitems(idx)
        if idx >= len(self.samples):
            raise AssertionError("sample index is out-of-range")
        sample = {key: self._unpack(args["dset"], idx, args["dtype"], args["shape"],
                                    args["compr_type"], **args["compr_kwargs"]) for key, args in self.target_args.items()}
        if self.transforms:
            sample = self.transforms(sample)
        return sample

    def close(self):
        """Closes the internal HDF5 file."""
        # note: if we dont do it explicitly, it will be done by the garbage collector on destruction, but it might take time...
        self.archive.close()


class ClassificationDataset(Dataset):
    """Classification dataset specialization interface.

    This specialization receives some extra parameters in its constructor and automatically defines
    a :class:`thelper.tasks.classif.Classification` task based on those. The derived class must still
    implement :func:`thelper.data.parsers.ClassificationDataset.__getitem__`, and it must still store its
    samples as dictionaries in ``self.samples`` to behave properly.

    .. seealso::
        | :class:`thelper.data.parsers.Dataset`
    """

    def __init__(self, class_names, input_key, label_key, meta_keys=None, transforms=None, deepcopy=False):
        """Classification dataset parser constructor.

        This constructor receives all extra arguments necessary to build a classification task object.

        Args:
            class_names: list of all class names (or labels) that will be associated with the samples.
            input_key: key used to index the input data in the loaded samples.
            label_key: key used to index the label (or class name) in the loaded samples.
            meta_keys: list of extra keys that will be available in the loaded samples.
            transforms: function or object that should be applied to all loaded samples in order to
                return the data in the requested transformed/augmented state.
            deepcopy: specifies whether this dataset interface should be deep-copied inside
                :class:`thelper.data.loaders.LoaderFactory` so that it may be shared between
                different threads. This is false by default, as we assume datasets do not contain a state
                or buffer that might cause problems in multi-threaded data loaders.
        """
        super(ClassificationDataset, self).__init__(transforms=transforms, deepcopy=deepcopy)
        self.task = thelper.tasks.Classification(class_names, input_key, label_key, meta_keys=meta_keys)

    @abstractmethod
    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        raise NotImplementedError


class SegmentationDataset(Dataset):
    """Segmentation dataset specialization interface.

    This specialization receives some extra parameters in its constructor and automatically defines
    its task (:class:`thelper.tasks.segm.Segmentation`) based on those. The derived class must still
    implement :func:`thelper.data.parsers.SegmentationDataset.__getitem__`, and it must still store its
    samples as dictionaries in ``self.samples`` to behave properly.

    .. seealso::
        | :class:`thelper.data.parsers.Dataset`
    """

    def __init__(self, class_names, input_key, label_map_key, meta_keys=None, dontcare=None, transforms=None, deepcopy=False):
        """Segmentation dataset parser constructor.

        This constructor receives all extra arguments necessary to build a segmentation task object.

        Args:
            class_names: list of all class names (or labels) that must be predicted in the image.
            input_key: key used to index the input image in the loaded samples.
            label_map_key: key used to index the label map in the loaded samples.
            meta_keys: list of extra keys that will be available in the loaded samples.
            transforms: function or object that should be applied to all loaded samples in order to
                return the data in the requested transformed/augmented state.
            deepcopy: specifies whether this dataset interface should be deep-copied inside
                :class:`thelper.data.loaders.LoaderFactory` so that it may be shared between
                different threads. This is false by default, as we assume datasets do not contain a state
                or buffer that might cause problems in multi-threaded data loaders.
        """
        super(SegmentationDataset, self).__init__(transforms=transforms, deepcopy=deepcopy)
        self.task = thelper.tasks.Segmentation(class_names, input_key, label_map_key, meta_keys=meta_keys, dontcare=dontcare)

    @abstractmethod
    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        raise NotImplementedError


class ImageDataset(Dataset):
    """Image dataset specialization interface.

    This specialization is used to parse simple image folders, and it does not fulfill the requirements of any
    specialized task constructors due to the lack of groundtruth data support. Therefore, it returns a basic task
    object (:class:`thelper.tasks.utils.Task`) with no set value for the groundtruth key, and it cannot be used to
    directly train a model. It can however be useful when simply visualizing, annotating, or testing raw data
    from a simple directory structure.

    .. seealso::
        | :class:`thelper.data.parsers.Dataset`
    """

    def __init__(self, root, transforms=None, image_key="image", path_key="path", idx_key="idx"):
        """Image dataset parser constructor.

        This constructor exposes some of the configurable keys used to index sample dictionaries.
        """
        super(ImageDataset, self).__init__(transforms=transforms)
        self.root = root
        if self.root is None or not os.path.isdir(self.root):
            raise AssertionError("invalid input data root '%s'" % self.root)
        self.image_key = image_key
        self.path_key = path_key
        self.idx_key = idx_key
        self.samples = []
        for folder, subfolder, files in os.walk(self.root):
            for file in files:
                ext = os.path.splitext(file)[1].lower()
                if ext in [".jpg", ".jpeg", ".bmp", ".png", ".ppm", ".pgm", ".tif"]:
                    self.samples.append({self.path_key: os.path.join(folder, file)})
        self.task = thelper.tasks.Task(self.image_key, None, [self.path_key, self.idx_key])

    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        if isinstance(idx, slice):
            return self._getitems(idx)
        if idx >= len(self.samples):
            raise AssertionError("sample index is out-of-range")
        if idx < 0:
            idx = len(self.samples) + idx
        sample = self.samples[idx]
        image_path = sample[self.path_key]
        image = cv.imread(image_path)
        if image is None:
            raise AssertionError("invalid image at '%s'" % image_path)
        sample = {
            self.image_key: image,
            self.idx_key: idx,
            **sample
        }
        if self.transforms:
            sample = self.transforms(sample)
        return sample


class ImageFolderDataset(ClassificationDataset):
    """Image folder dataset specialization interface for classification tasks.

    This specialization is used to parse simple image subfolders, and it essentially replaces the very
    basic ``torchvision.datasets.ImageFolder`` interface with similar functionalities. It it used to provide
    a proper task interface as well as path metadata in each loaded packet for metrics/logging output.

    .. seealso::
        | :class:`thelper.data.parsers.ImageDataset`
        | :class:`thelper.data.parsers.ClassificationDataset`
    """

    def __init__(self, root, transforms=None, image_key="image", label_key="label", path_key="path", idx_key="idx"):
        """Image folder dataset parser constructor."""
        self.root = root
        if self.root is None or not os.path.isdir(self.root):
            raise AssertionError("invalid input data root '%s'" % self.root)
        class_map = {}
        for child in os.listdir(self.root):
            if os.path.isdir(os.path.join(self.root, child)):
                class_map[child] = []
        if not class_map:
            raise AssertionError("could not find any image folders at '%s'" % self.root)
        image_exts = [".jpg", ".jpeg", ".bmp", ".png", ".ppm", ".pgm", ".tif"]
        self.image_key = image_key
        self.path_key = path_key
        self.idx_key = idx_key
        self.label_key = label_key
        samples = []
        for class_name in class_map:
            class_folder = os.path.join(self.root, class_name)
            for folder, subfolder, files in os.walk(class_folder):
                for file in files:
                    ext = os.path.splitext(file)[1].lower()
                    if ext in image_exts:
                        class_map[class_name].append(len(samples))
                        samples.append({
                            self.path_key: os.path.join(folder, file),
                            self.label_key: class_name
                        })
        class_map = {k: v for k, v in class_map.items() if len(v) > 0}
        if not class_map:
            raise AssertionError("could not locate any subdir in '%s' with images to load" % self.root)
        meta_keys = [self.path_key, self.idx_key]
        super(ImageFolderDataset, self).__init__(class_names=list(class_map.keys()), input_key=self.image_key,
                                                 label_key=self.label_key, meta_keys=meta_keys, transforms=transforms)
        self.samples = samples

    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        if isinstance(idx, slice):
            return self._getitems(idx)
        if idx >= len(self.samples):
            raise AssertionError("sample index is out-of-range")
        if idx < 0:
            idx = len(self.samples) + idx
        sample = self.samples[idx]
        image_path = sample[self.path_key]
        image = cv.imread(image_path)
        if image is None:
            raise AssertionError("invalid image at '%s'" % image_path)
        sample = {
            self.image_key: image,
            self.idx_key: idx,
            **sample
        }
        if self.transforms:
            sample = self.transforms(sample)
        return sample


class SuperResFolderDataset(Dataset):
    """Image folder dataset specialization interface for super-resolution tasks.

    This specialization is used to parse simple image subfolders, and it essentially replaces the very
    basic ``torchvision.datasets.ImageFolder`` interface with similar functionalities. It it used to provide
    a proper task interface as well as path/class metadata in each loaded packet for metrics/logging output.
    """

    def __init__(self, root, downscale_factor=2.0, rescale_lowres=True, center_crop=None, transforms=None,
                 lowres_image_key="lowres_image", highres_image_key="highres_image", path_key="path", idx_key="idx", label_key="label"):
        """Image folder dataset parser constructor."""
        if isinstance(downscale_factor, int):
            downscale_factor = float(downscale_factor)
        if not isinstance(downscale_factor, float) or downscale_factor <= 1.0:
            raise AssertionError("invalid downscale factor (should be greater than one)")
        self.downscale_factor = downscale_factor
        self.rescale_lowres = rescale_lowres
        if center_crop is not None:
            if isinstance(center_crop, int):
                center_crop = (center_crop, center_crop)
            if not isinstance(center_crop, (list, tuple)):
                raise AssertionError("invalid center crop size type")
        self.center_crop = center_crop
        self.root = root
        if self.root is None or not os.path.isdir(self.root):
            raise AssertionError("invalid input data root '%s'" % self.root)
        class_map = {}
        for child in os.listdir(self.root):
            if os.path.isdir(os.path.join(self.root, child)):
                class_map[child] = []
        if not class_map:
            raise AssertionError("could not find any image folders at '%s'" % self.root)
        image_exts = [".jpg", ".jpeg", ".bmp", ".png", ".ppm", ".pgm", ".tif"]
        self.lowres_image_key = lowres_image_key
        self.highres_image_key = highres_image_key
        self.path_key = path_key
        self.idx_key = idx_key
        self.label_key = label_key  # to provide folder names
        samples = []
        for class_name in class_map:
            class_folder = os.path.join(self.root, class_name)
            for folder, subfolder, files in os.walk(class_folder):
                for file in files:
                    ext = os.path.splitext(file)[1].lower()
                    if ext in image_exts:
                        class_map[class_name].append(len(samples))
                        samples.append({
                            self.path_key: os.path.join(folder, file),
                            self.label_key: class_name
                        })
        class_map = {k: v for k, v in class_map.items() if len(v) > 0}
        if not class_map:
            raise AssertionError("could not locate any subdir in '%s' with images to load" % self.root)
        meta_keys = [self.path_key, self.idx_key, self.label_key]
        super(SuperResFolderDataset, self).__init__(transforms=transforms)
        self.task = thelper.tasks.SuperResolution(input_key=self.lowres_image_key, target_key=self.highres_image_key, meta_keys=meta_keys)
        self.samples = samples

    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        if isinstance(idx, slice):
            return self._getitems(idx)
        if idx >= len(self.samples):
            raise AssertionError("sample index is out-of-range")
        if idx < 0:
            idx = len(self.samples) + idx
        sample = self.samples[idx]
        image_path = sample[self.path_key]
        image = cv.imread(image_path)
        if image is None:
            raise AssertionError("invalid image at '%s'" % image_path)
        if self.center_crop is not None:
            tl = (image.shape[1] // 2 - self.center_crop[0] // 2,
                  image.shape[0] // 2 - self.center_crop[1] // 2)
            br = (tl[0] + self.center_crop[0], tl[1] + self.center_crop[1])
            image = thelper.draw.safe_crop(image, tl, br)
        scale = 1.0 / self.downscale_factor
        image_lowres = cv.resize(image, dsize=(0, 0), fx=scale, fy=scale)
        if self.rescale_lowres:
            image_lowres = cv.resize(image_lowres, dsize=(image.shape[1], image.shape[0]))
        sample = {
            self.lowres_image_key: image_lowres,
            self.highres_image_key: image,
            self.idx_key: idx,
            **sample
        }
        if self.transforms:
            sample = self.transforms(sample)
        return sample


class ExternalDataset(Dataset):
    """External dataset interface.

    This interface allows external classes to be instantiated automatically in the framework through
    a configuration file, as long as they themselves provide implementations for  ``__getitem__`` and
    ``__len__``. This includes all derived classes of ``torch.utils.data.Dataset`` such as
    ``torchvision.datasets.ImageFolder``, and the specialized versions such as ``torchvision.datasets.CIFAR10``.

    Note that for this interface to be compatible with our runtime instantiation rules, the constructor
    needs to receive a fully constructed task object. This object is currently constructed in
    :func:`thelper.data.utils.create_parsers` based on extra parameters; see the code there for more
    information.

    Attributes:
        dataset_type: type of the internally instantiated or provided dataset object.
        warned_dictionary: specifies whether the user was warned about missing keys in the output
            samples dictionaries.

    .. seealso::
        | :class:`thelper.data.parsers.Dataset`
    """

    def __init__(self, dataset, task, transforms=None, deepcopy=False, **kwargs):
        """External dataset parser constructor.

        Args:
            dataset: fully qualified name of the dataset object to instantiate, or the dataset itself.
            task: fully constructed task object providing key information for sample loading.
            transforms: function or object that should be applied to all loaded samples in order to
                return the data in the requested transformed/augmented state.
            deepcopy: specifies whether this dataset interface should be deep-copied inside
                :class:`thelper.data.loaders.LoaderFactory` so that it may be shared between
                different threads. This is false by default, as we assume datasets do not contain a state
                or buffer that might cause problems in multi-threaded data loaders.
        """
        super(ExternalDataset, self).__init__(transforms=transforms, deepcopy=deepcopy)
        if isinstance(dataset, str):
            dataset = thelper.utils.import_class(dataset)
        if dataset is None or not hasattr(dataset, "__getitem__") or not hasattr(dataset, "__len__"):
            raise AssertionError("external dataset type must implement '__getitem__' and '__len__' methods")
        if inspect.isclass(dataset):
            self.samples = dataset(**kwargs)
        else:
            self.samples = dataset
        if task is None or not isinstance(task, thelper.tasks.Task):
            raise AssertionError("task type must derive from thelper.tasks.Task")
        self.dataset_type = type(self.samples)
        self.task = task
        self.warned_dictionary = False

    def _get_derived_name(self):
        """Returns a pretty-print version of the external class's name."""
        return self.dataset_type.__module__ + "." + self.dataset_type.__qualname__

    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        if isinstance(idx, slice):
            return self._getitems(idx)
        if idx >= len(self.samples):
            raise AssertionError("sample index is out-of-range")
        sample = self.samples[idx]
        if sample is None:
            # since might have provided an invalid sample count before, it's dangerous to skip empty samples here
            raise AssertionError("invalid sample received in external dataset impl")
        warn_dictionary = False
        if isinstance(sample, (list, tuple)):
            out_sample_list = []
            for key_idx, subs in enumerate(sample):
                if isinstance(subs, PIL.Image.Image):
                    subs = np.array(subs)
                out_sample_list.append(subs)
            sample = {str(key_idx): out_sample_list[key_idx] for key_idx in range(len(out_sample_list))}
            warn_dictionary = True
        elif isinstance(sample, (np.ndarray, PIL.Image.Image, torch.Tensor)):
            sample = {"0": sample}
            warn_dictionary = True
        if warn_dictionary and not self.warned_dictionary:
            logger.warning("dataset '%s' not returning samples as dictionaries;"
                           " will blindly map elements to their indices" % self._get_derived_name())
            self.warned_dictionary = True
        if self.transforms:
            sample = self.transforms(sample)
        return sample



import glob
import os

import numpy as np

import thelper

# note: the bbox data used here is taken from the 'Object Detection Metrics' repository of Rafael Padilla.
# See https://github.com/rafaelpadilla/Object-Detection-Metrics for more information.
# The original code is distributed under the MIT License, Copyright (c) 2018 Rafael Padilla.


def get_bboxes(dirpath, gt=False):
    bboxes = []
    files = glob.glob(os.path.join(dirpath, "*.txt"))
    files.sort()
    for filename in files:
        with open(filename, "r") as fd:
            for line in fd:
                line = line.replace("\n", "")
                if line.replace(" ", "") == "":
                    continue
                split = line.split(" ")
                if gt:
                    xmin = int(split[1])
                    ymin = int(split[2])
                    xmax = xmin + int(split[3])
                    ymax = ymin + int(split[4])
                    bboxes.append(thelper.tasks.detect.BoundingBox(split[0],
                                                                   [xmin, ymin, xmax, ymax],
                                                                   image_id=os.path.basename(filename)))
                else:
                    xmin = int(split[2])
                    ymin = int(split[3])
                    xmax = xmin + int(split[4])
                    ymax = ymin + int(split[5])
                    bboxes.append(thelper.tasks.detect.BoundingBox(split[0],
                                                                   [xmin, ymin, xmax, ymax],
                                                                   confidence=float(split[1]),
                                                                   image_id=os.path.basename(filename)))
    return bboxes


def test_bbox_map():
    curr_path = os.path.dirname(os.path.abspath(__file__))
    preds_path = os.path.join(curr_path, "detections")
    assert os.path.isdir(preds_path)
    targets_path = os.path.join(curr_path, "groundtruths")
    assert os.path.isdir(targets_path)
    preds = get_bboxes(preds_path, False)
    targets = get_bboxes(targets_path, True)
    assert targets and preds
    task = thelper.tasks.Detection(["person"], "in", "gt")
    res = thelper.optim.compute_pascalvoc_metrics(preds, targets, task, iou_threshold=0.3)
    ap = res["person"]["AP"]
    assert np.isclose(ap, 0.24568668046928915)  # obtained via the original example



"""General utilities module.

This module only contains non-ML specific functions, i/o helpers,
and matplotlib/pyplot drawing calls.
"""
import copy
import errno
import functools
import glob
import hashlib
import importlib
import importlib.util
import inspect
import io
import json
import logging
import os
import pathlib
import pickle
import platform
import re
import sys
import time
from typing import TYPE_CHECKING

import numpy as np
import torch
import yaml

import thelper.typedefs  # noqa: F401

if TYPE_CHECKING:
    from typing import Any, List, Optional, Type  # noqa: F401
    from types import FunctionType  # noqa: F401

logger = logging.getLogger(__name__)
bypass_queries = False
warned_generic_draw = False
fixed_yaml_parsing = False


class Struct:
    """Generic runtime-defined C-like data structure (maps constructor elements to fields)."""

    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)

    def __repr__(self):
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            "(" + ", ".join([f"{key}={repr(val)}" for key, val in self.__dict__.items()]) + ")"


def test_cuda_device_availability(device_idx):
    # type: (int) -> bool
    """Tests the availability of a single cuda device and returns its status."""
    # noinspection PyBroadException
    try:
        torch.cuda.set_device(device_idx)
        test_val = torch.cuda.FloatTensor([1])
        return test_val.cpu().item() == 1.0
    except Exception:
        return False


def get_available_cuda_devices(attempts_per_device=5):
    # type: (Optional[int]) -> List[int]
    """
    Tests all visible cuda devices and returns a list of available ones.

    Returns:
        List of available cuda device IDs (integers). An empty list means no
        cuda device is available, and the app should fallback to cpu.
    """
    if not torch.cuda.is_available() or torch.cuda.device_count() == 0:
        return []
    devices_available = [False] * torch.cuda.device_count()
    attempt_broadcast = False
    for attempt in range(attempts_per_device):
        for device_id in range(torch.cuda.device_count()):
            if not devices_available[device_id]:
                if not attempt_broadcast:
                    logger.debug("testing availability of cuda device #%d (%s)" % (
                        device_id, torch.cuda.get_device_name(device_id)
                    ))
                devices_available[device_id] = test_cuda_device_availability(device_id)
        attempt_broadcast = True
    return [device_id for device_id, available in enumerate(devices_available) if available]


def setup_plt(config):
    """Parses the provided config for matplotlib flags and sets up its global state accordingly."""
    import matplotlib.pyplot as plt
    config = get_key_def(["plt", "pyplot", "matplotlib"], config, {})
    if "backend" in config:
        import matplotlib
        matplotlib.use(get_key("backend", config))
    plt.interactive(get_key_def("interactive", config, False))


# noinspection PyUnusedLocal
def setup_cv2(config):
    """Parses the provided config for OpenCV flags and sets up its global state accordingly."""
    # https://github.com/pytorch/pytorch/issues/1355
    import cv2 as cv
    cv.setNumThreads(0)
    cv.ocl.setUseOpenCL(False)
    # todo: add more global opencv flags setups here


def setup_gdal(config):
    """Parses the provided config for GDAL flags and sets up its global state accordingly."""
    config = get_key_def("gdal", config, {})
    if "proj_search_path" in config:
        import osr
        osr.SetPROJSearchPath(config["proj_search_path"])


def setup_cudnn(config):
    """Parses the provided config for CUDNN flags and sets up PyTorch accordingly."""
    if "cudnn" in config and isinstance(config["cudnn"], dict):
        config = config["cudnn"]
        if "benchmark" in config:
            cudnn_benchmark_flag = str2bool(config["benchmark"])
            logger.debug("cudnn benchmark mode = %s" % str(cudnn_benchmark_flag))
            torch.backends.cudnn.benchmark = cudnn_benchmark_flag
        if "deterministic" in config:
            cudnn_deterministic_flag = str2bool(config["deterministic"])
            logger.debug("cudnn deterministic mode = %s" % str(cudnn_deterministic_flag))
            torch.backends.cudnn.deterministic = cudnn_deterministic_flag
    else:
        if "cudnn_benchmark" in config:
            cudnn_benchmark_flag = str2bool(config["cudnn_benchmark"])
            logger.debug("cudnn benchmark mode = %s" % str(cudnn_benchmark_flag))
            torch.backends.cudnn.benchmark = cudnn_benchmark_flag
        if "cudnn_deterministic" in config:
            cudnn_deterministic_flag = str2bool(config["cudnn_deterministic"])
            logger.debug("cudnn deterministic mode = %s" % str(cudnn_deterministic_flag))
            torch.backends.cudnn.deterministic = cudnn_deterministic_flag


def setup_globals(config):
    """Parses the provided config for global flags and sets up the global state accordingly."""
    if "bypass_queries" in config and config["bypass_queries"]:
        global bypass_queries
        bypass_queries = True
    setup_plt(config)
    setup_cv2(config)
    setup_gdal(config)
    setup_cudnn(config)


def load_checkpoint(ckpt,                      # type: thelper.typedefs.CheckpointLoadingType
                    map_location=None,         # type: Optional[thelper.typedefs.MapLocationType]
                    always_load_latest=False,  # type: Optional[bool]
                    check_version=True,        # type: Optional[bool]
                    ):                         # type: (...) -> thelper.typedefs.CheckpointContentType
    """Loads a session checkpoint via PyTorch, check its compatibility, and returns its data.

    If the ``ckpt`` parameter is a path to a valid directory, then that directly will be searched for
    a checkpoint. If multiple checkpoints are found, the latest will be returned (based on the epoch
    index in its name). iF ``always_load_latest`` is set to False and if a checkpoint named
    ``ckpt.best.pth`` is found, it will be returned instead.

    Args:
        ckpt: a file-like object or a path to the checkpoint file or session directory.
        map_location: a function, string or a dict specifying how to remap storage
            locations. See ``torch.load`` for more information.
        always_load_latest: toggles whether to always try to load the latest checkpoint
            if a session directory is provided (instead of loading the 'best' checkpoint).
        check_version: toggles whether the checkpoint's version should be checked for
            compatibility issues, and query the user for how to proceed.

    Returns:
        Content of the checkpoint (a dictionary).
    """
    if map_location is None and not get_available_cuda_devices():
        map_location = 'cpu'
    if isinstance(ckpt, str) and os.path.isdir(ckpt):
        logger.debug("will search directory '%s' for a checkpoint to load..." % ckpt)
        search_ckpt_dir = os.path.join(ckpt, "checkpoints")
        if os.path.isdir(search_ckpt_dir):
            search_dir = search_ckpt_dir
        else:
            search_dir = ckpt
        ckpt_paths = glob.glob(os.path.join(search_dir, "ckpt.*.pth"))
        if not ckpt_paths:
            raise AssertionError("could not find any valid checkpoint files in directory '%s'" % search_dir)
        latest_epoch, latest_day, latest_time = -1, -1, -1
        for ckpt_path in ckpt_paths:
            # note: the 2nd field in the name should be the epoch index, or 'best' if final checkpoint
            split = os.path.basename(ckpt_path).split(".")
            tag = split[1]
            if tag == "best" and (not always_load_latest or latest_epoch == -1):
                # if eval-only, always pick the best checkpoint; otherwise, only pick if nothing else exists
                ckpt = ckpt_path
                if not always_load_latest:
                    break
            elif tag != "best":
                log_stamp = split[2] if len(split) > 2 else ""
                log_stamp = "fake-0-0" if log_stamp.count("-") != 2 else log_stamp
                epoch_stamp, day_stamp, time_stamp = int(tag), int(log_stamp.split("-")[1]), int(log_stamp.split("-")[2])
                if epoch_stamp > latest_epoch or day_stamp > latest_day or time_stamp > latest_time:
                    ckpt, latest_epoch, latest_day, latest_time = ckpt_path, epoch_stamp, day_stamp, time_stamp
        if not os.path.isfile(ckpt):
            raise AssertionError("could not find valid checkpoint at '%s'" % ckpt)
    basepath = None
    if isinstance(ckpt, str):
        logger.debug("parsing checkpoint at '%s'" % ckpt)
        basepath = os.path.dirname(os.path.abspath(ckpt))
    else:
        if hasattr(ckpt, "name"):
            logger.debug("parsing checkpoint provided via file object")
            basepath = os.path.dirname(os.path.abspath(ckpt.name))
    ckptdata = torch.load(ckpt, map_location=map_location)
    if not isinstance(ckptdata, dict):
        raise AssertionError("unexpected checkpoint data type")
    if check_version:
        good_version = False
        from thelper import __version__ as curr_ver
        if "version" not in ckptdata:
            logger.warning("checkpoint missing internal version tag")
            ckpt_ver_str = "0.0.0"
        else:
            ckpt_ver_str = ckptdata["version"]
            if not isinstance(ckpt_ver_str, str) or len(ckpt_ver_str.split(".")) != 3:
                raise AssertionError("unexpected checkpoint version formatting")
            # by default, checkpoints should be from the same minor version, we warn otherwise
            versions = [curr_ver.split("."), ckpt_ver_str.split(".")]
            if versions[0][0] != versions[1][0]:
                logger.error("incompatible checkpoint, major version mismatch (%s vs %s)" % (curr_ver, ckpt_ver_str))
            elif versions[0][1] != versions[1][1]:
                logger.warning("outdated checkpoint, minor version mismatch (%s vs %s)" % (curr_ver, ckpt_ver_str))
            else:
                good_version = True
        if not good_version:
            answer = query_string("Checkpoint version unsupported (framework=%s, checkpoint=%s); how do you want to proceed?" %
                                  (curr_ver, ckpt_ver_str), choices=["continue", "migrate", "abort"], default="migrate", bypass="migrate")
            if answer == "abort":
                logger.error("checkpoint out-of-date; user aborted")
                sys.exit(1)
            elif answer == "continue":
                logger.warning("will attempt to load checkpoint anyway (might crash later due to incompatibilities)")
            elif answer == "migrate":
                ckptdata = migrate_checkpoint(ckptdata)
    # load model trace if needed (we do it here since we can locate the neighboring file)
    if "model" in ckptdata and isinstance(ckptdata["model"], str):
        trace_path = None
        if os.path.isfile(ckptdata["model"]):
            trace_path = ckptdata["model"]
        elif basepath is not None and os.path.isfile(os.path.join(basepath, ckptdata["model"])):
            trace_path = os.path.join(basepath, ckptdata["model"])
        if trace_path is not None:
            if trace_path.endswith(".pth"):
                ckptdata["model"] = torch.load(trace_path, map_location=map_location)
            elif trace_path.endswith(".zip"):
                ckptdata["model"] = torch.jit.load(trace_path, map_location=map_location)
    return ckptdata


def migrate_checkpoint(ckptdata,  # type: thelper.typedefs.CheckpointContentType
                       ):         # type: (...) -> thelper.typedefs.CheckpointContentType
    """Migrates the content of an incompatible or outdated checkpoint to the current version of the framework.

    This function might not be able to fix all backward compatibility issues (e.g. it cannot fix class interfaces
    that were changed). Perfect reproductibility of tests cannot be guaranteed either if this migration tool is used.

    Args:
        ckptdata: checkpoint data in dictionary form obtained via ``thelper.utils.load_checkpoint``. Note that
            the data contained in this dictionary will be modified in-place.

    Returns:
        An updated checkpoint dictionary that should be compatible with the current version of the framework.
    """
    if not isinstance(ckptdata, dict):
        raise AssertionError("unexpected ckptdata type")
    from thelper import __version__ as curr_ver
    curr_ver = [int(num) for num in curr_ver.split(".")]
    ckpt_ver_str = ckptdata["version"] if "version" in ckptdata else "0.0.0"
    ckpt_ver = [int(num) for num in ckpt_ver_str.split(".")]
    if (ckpt_ver[0] > curr_ver[0] or (ckpt_ver[0] == curr_ver[0] and ckpt_ver[1] > curr_ver[1]) or
       (ckpt_ver[0:2] == curr_ver[0:2] and ckpt_ver[2] > curr_ver[2])):
        raise AssertionError("cannot migrate checkpoints from future versions!")
    if "config" not in ckptdata:
        raise AssertionError("checkpoint migration requires config")
    old_config = ckptdata["config"]
    new_config = migrate_config(copy.deepcopy(old_config), ckpt_ver_str)
    if ckpt_ver == [0, 0, 0]:
        logger.warning("trying to migrate checkpoint data from v0.0.0; all bets are off")
    else:
        logger.info("trying to migrate checkpoint data from v%s" % ckpt_ver_str)
    if ckpt_ver[0] <= 0 and ckpt_ver[1] <= 1:
        # combine 'host' and 'time' fields into 'source'
        if "host" in ckptdata and "time" in ckptdata:
            ckptdata["source"] = ckptdata["host"] + ckptdata["time"]
            del ckptdata["host"]
            del ckptdata["time"]
        # update classif task interface
        if "task" in ckptdata and isinstance(ckptdata["task"], thelper.tasks.classif.Classification):
            ckptdata["task"] = str(thelper.tasks.classif.Classification(class_names=ckptdata["task"].class_names,
                                                                        input_key=ckptdata["task"].input_key,
                                                                        label_key=ckptdata["task"].label_key,
                                                                        meta_keys=ckptdata["task"].meta_keys))
        # move 'state_dict' field to 'model'
        if "state_dict" in ckptdata:
            ckptdata["model"] = ckptdata["state_dict"]
            del ckptdata["state_dict"]
        # create 'model_type' and 'model_params' fields
        if "model" in new_config:
            if "type" in new_config["model"]:
                ckptdata["model_type"] = new_config["model"]["type"]
            else:
                ckptdata["model_type"] = None
            if "params" in new_config["model"]:
                ckptdata["model_params"] = copy.deepcopy(new_config["model"]["params"])
            else:
                ckptdata["model_params"] = {}
        # TODO: create 'scheduler' field to restore previous state? (not so important for early versions)
        # ckpt_ver = [0, 2, 0]  # set ver for next update step
    # if ckpt_ver[0] <= x and ckpt_ver[1] <= y and ckpt_ver[2] <= z:
    #     ... add more compatibility fixes here
    ckptdata["config"] = new_config
    return ckptdata


def migrate_config(config,        # type: thelper.typedefs.ConfigDict
                   cfg_ver_str,   # type: str
                   ):             # type: (...) -> thelper.typedefs.ConfigDict
    """Migrates the content of an incompatible or outdated configuration to the current version of the framework.

    This function might not be able to fix all backward compatibility issues (e.g. it cannot fix class interfaces
    that were changed). Perfect reproducibility of tests cannot be guaranteed either if this migration tool is used.

    Args:
        config: session configuration dictionary obtained e.g. by parsing a JSON file. Note that the data contained
            in this dictionary will be modified in-place.
        cfg_ver_str: string representing the version for which the configuration was created (e.g. "0.2.0").

    Returns:
        An updated configuration dictionary that should be compatible with the current version of the framework.
    """
    if not isinstance(config, dict):
        raise AssertionError("unexpected config type")
    if not isinstance(cfg_ver_str, str) or len(cfg_ver_str.split(".")) != 3:
        raise AssertionError("unexpected checkpoint version formatting")
    from thelper import __version__ as curr_ver
    curr_ver = [int(num) for num in curr_ver.split(".")]
    cfg_ver = [int(num) for num in cfg_ver_str.split(".")]
    if (cfg_ver[0] > curr_ver[0] or (cfg_ver[0] == curr_ver[0] and cfg_ver[1] > curr_ver[1]) or
       (cfg_ver[0:2] == curr_ver[0:2] and cfg_ver[2] > curr_ver[2])):
        raise AssertionError("cannot migrate configs from future versions!")
    if cfg_ver == [0, 0, 0]:
        logger.warning("trying to migrate config from v0.0.0; all bets are off")
    else:
        logger.info("trying to migrate config from v%s" % cfg_ver_str)
    if cfg_ver[0] <= 0 and cfg_ver[1] < 1:
        # must search for name-value parameter lists and convert them to dictionaries
        def name_value_replacer(cfg):
            if isinstance(cfg, dict):
                for key, val in cfg.items():
                    if (key == "params" or key == "parameters") and isinstance(val, list) and \
                       all([isinstance(p, dict) and list(p.keys()) == ["name", "value"] for p in val]):
                        cfg["params"] = {param["name"]: name_value_replacer(param["value"]) for param in val}
                        if key == "parameters":
                            del cfg["parameters"]
                    elif isinstance(val, (dict, list)):
                        cfg[key] = name_value_replacer(val)
            elif isinstance(cfg, list):
                for idx, val in enumerate(cfg):
                    cfg[idx] = name_value_replacer(val)
            return cfg
        config = name_value_replacer(config)
        # must replace "data_config" section by "loaders"
        if "data_config" in config:
            config["loaders"] = config["data_config"]
            del config["data_config"]
        # remove deprecated name attribute for models
        if "model" in config and isinstance(config["model"], dict) and "name" in config["model"]:
            del config["model"]["name"]
        # must update import targets wrt class name refactorings
        def import_refactoring(cfg):  # noqa: E306
            if isinstance(cfg, dict):
                for key, val in cfg.items():
                    cfg[key] = import_refactoring(val)
            elif isinstance(cfg, list):
                for idx, val in enumerate(cfg):
                    cfg[idx] = import_refactoring(val)
            elif isinstance(cfg, str) and cfg.startswith("thelper."):
                cfg = thelper.utils.resolve_import(cfg)
            return cfg
        config = import_refactoring(config)
        if "trainer" in config and isinstance(config["trainer"], dict):
            trainer_cfg = config["trainer"]    # type: thelper.typedefs.ConfigDict
            # move 'loss' section to 'optimization' section
            if "loss" in trainer_cfg:
                if "optimization" not in trainer_cfg or not isinstance(trainer_cfg["optimization"], dict):
                    trainer_cfg["optimization"] = {}
                trainer_cfg["optimization"]["loss"] = trainer_cfg["loss"]
                del trainer_cfg["loss"]
            # replace all devices with cuda:all
            if "train_device" in trainer_cfg:
                del trainer_cfg["train_device"]
            if "valid_device" in trainer_cfg:
                del trainer_cfg["valid_device"]
            if "test_device" in trainer_cfg:
                del trainer_cfg["test_device"]
            if "device" not in trainer_cfg:
                trainer_cfg["device"] = "cuda:all"
            # remove params from trainer config
            if "params" in trainer_cfg:
                if not isinstance(trainer_cfg["params"], (dict, list)) or trainer_cfg["params"]:
                    logger.warning("removing non-empty parameter section from trainer config")
                del trainer_cfg["params"]
        cfg_ver = [0, 1, 0]  # set ver for next update step
    if cfg_ver[0] <= 0 and cfg_ver[1] <= 1:
        # remove 'force_convert' flags from all transform pipelines + build augment pipeline wrappers
        def remove_force_convert(cfg):  # noqa: E306
            if isinstance(cfg, list):
                for idx, stage in enumerate(cfg):
                    cfg[idx] = remove_force_convert(stage)
            elif isinstance(cfg, dict):
                if "parameters" in cfg:
                    cfg["params"] = cfg["parameters"]
                    del cfg["parameters"]
                if "operation" in cfg and cfg["operation"] == "thelper.transforms.TransformWrapper":
                    if "params" in cfg and "force_convert" in cfg["params"]:
                        del cfg["params"]["force_convert"]
                for key, stage in cfg.items():
                    cfg[key] = remove_force_convert(stage)
            return cfg
        for pipeline in ["base_transforms", "train_augments", "valid_augments", "test_augments"]:
            if "loaders" in config and isinstance(config["loaders"], dict) and pipeline in config["loaders"]:
                if pipeline.endswith("_augments"):
                    stages = config["loaders"][pipeline]
                    for stage in stages:
                        if "append" in stage:
                            if stage["append"]:
                                logger.warning("overriding augmentation stage ordering")
                            del stage["append"]
                        if "operation" in stage and stage["operation"] == "Augmentor.Pipeline":
                            if "params" in stage:
                                stage["params"] = stage["params"]["operations"]
                            elif "parameters" in stage:
                                stage["params"] = stage["parameters"]["operations"]
                                del stage["parameters"]
                    config["loaders"][pipeline] = {"append": False, "transforms": remove_force_convert(stages)}
                else:
                    config["loaders"][pipeline] = remove_force_convert(config["loaders"][pipeline])
        cfg_ver = [0, 2, 0]  # set ver for next update step
    if cfg_ver[0] <= 0 and cfg_ver[1] <= 2 and cfg_ver[2] < 5:
        # TODO: add scheduler 0-based step fix here? (unlikely to cause serious issues)
        cfg_ver = [0, 2, 5]  # set ver for next update step
    if cfg_ver[0] <= 0 and cfg_ver[1] <= 3 and cfg_ver[2] < 6:
        if "trainer" in config:
            if "eval_metrics" in config["trainer"]:
                assert "valid_metrics" not in config["trainer"]
                config["trainer"]["valid_metrics"] = config["trainer"]["eval_metrics"]
                del config["trainer"]["eval_metrics"]
            for mset in ["train_metrics", "valid_metrics", "test_metrics", "metrics"]:
                if mset in config["trainer"]:
                    metrics_config = config["trainer"][mset]
                    for mname, mcfg in metrics_config.items():
                        if "type" in mcfg and mcfg["type"].endswith("ExternalMetric"):
                            assert "params" in mcfg
                            assert "goal" in mcfg["params"]
                            mcfg["params"]["metric_goal"] = mcfg["params"]["goal"]
                            del mcfg["params"]["goal"]
                            if "metric_params" in mcfg["params"]:
                                if isinstance(mcfg["params"]["metric_params"], list):
                                    assert not mcfg["params"]["metric_params"], "cannot fill in kw names"
                                    mcfg["params"]["metric_params"] = {}
                        elif "type" in mcfg and mcfg["type"].endswith("ROCCurve"):
                            assert "params" in mcfg
                            if "log_params" in mcfg["params"]:
                                logger.warning("disabling logging via ROCCurve metric")
                                del mcfg["params"]["log_params"]
        cfg_ver = [0, 3, 6]  # set ver for next update step
    # if cfg_ver[0] <= x and cfg_ver[1] <= y and cfg_ver[2] <= z:
    #     ... add more compatibility fixes here
    return config


def download_file(url, root, filename, md5=None):
    """Downloads a file from a given URL to a local destination.

    Args:
        url: path to query for the file (query will be based on urllib).
        root: destination folder where the file should be saved.
        filename: destination name for the file.
        md5: optional, for md5 integrity check.

    Returns:
        The path to the downloaded file.
    """
    # inspired from torchvision.datasets.utils.download_url; no dep check
    from six.moves import urllib
    root = os.path.expanduser(root)
    fpath = os.path.join(root, filename)
    try:
        os.makedirs(root)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise
    if not os.path.isfile(fpath):
        logger.info("Downloading %s to %s ..." % (url, fpath))
        urllib.request.urlretrieve(url, fpath, reporthook)
        sys.stdout.write("\r")
        sys.stdout.flush()
    if md5 is not None:
        md5o = hashlib.md5()
        with open(fpath, 'rb') as f:
            for chunk in iter(lambda: f.read(1024 * 1024), b''):
                md5o.update(chunk)
        md5c = md5o.hexdigest()
        if md5c != md5:
            raise AssertionError("md5 check failed for '%s'" % fpath)
    return fpath


def extract_tar(filepath, root, flags="r:gz"):
    """Extracts the content of a tar file to a specific location.

    Args:
        filepath: location of the tar archive.
        root: where to extract the archive's content.
        flags: extra flags passed to ``tarfile.open``.
    """
    import tarfile

    class _FileWrapper(io.FileIO):
        def __init__(self, path, *args, **kwargs):
            self.start_time = time.time()
            self._size = os.path.getsize(path)
            super().__init__(path, *args, **kwargs)

        def read(self, *args, **kwargs):
            duration = time.time() - self.start_time
            progress_size = self.tell()
            speed = str(int(progress_size / (1024 * duration))) if duration > 0 else "?"
            percent = min(int(progress_size * 100 / self._size), 100)
            sys.stdout.write("\r\t=> extracted %d%% (%d MB) @ %s KB/s..." %
                             (percent, progress_size / (1024 * 1024), speed))
            sys.stdout.flush()
            return io.FileIO.read(self, *args, **kwargs)

    cwd = os.getcwd()
    tar = tarfile.open(fileobj=_FileWrapper(filepath), mode=flags)
    os.chdir(root)
    tar.extractall()
    tar.close()
    os.chdir(cwd)
    sys.stdout.write("\r")
    sys.stdout.flush()


def reporthook(count, block_size, total_size):
    """Report hook used to display a download progression bar when using urllib requests."""
    global start_time
    if count == 0:
        start_time = time.time()
        return
    duration = time.time() - start_time
    progress_size = int(count * block_size)
    speed = str(int(progress_size / (1024 * duration))) if duration > 0 else "?"
    percent = min(int(count * block_size * 100 / total_size), 100)
    sys.stdout.write("\r\t=> downloaded %d%% (%d MB) @ %s KB/s..." %
                     (percent, progress_size / (1024 * 1024), speed))
    sys.stdout.flush()


def init_logger(log_level=logging.NOTSET, filename=None, force_stdout=False):
    """Initializes the framework logger with a specific filter level, and optional file output."""
    if getattr(thelper, "LOGGER_INITIALIZED", None) is None:
        logging.getLogger().setLevel(logging.NOTSET)
        thelper.logger.propagate = 0
        logger_format = logging.Formatter("[%(asctime)s - %(name)s] %(levelname)s : %(message)s")
        if filename is not None:
            logger_fh = logging.FileHandler(filename)
            logger_fh.setLevel(logging.NOTSET)
            logger_fh.setFormatter(logger_format)
            thelper.logger.addHandler(logger_fh)
        stream = sys.stdout if force_stdout else None
        logger_ch = logging.StreamHandler(stream=stream)
        logger_ch.setLevel(log_level)
        logger_ch.setFormatter(logger_format)
        thelper.logger.addHandler(logger_ch)
        setattr(thelper, "LOGGER_INITIALIZED", True)


def resolve_import(fullname):
    # type: (str) -> str
    """
    Class name resolver.

    Takes a string corresponding to a module and class fullname to be imported with :func:`thelper.utils.import_class`
    and resolves any back compatibility issues related to renamed or moved classes.

    Args:
        fullname: the fully qualified class name to be resolved.

    Returns:
        The resolved class fullname.
    """
    removed_cases = [
        'thelper.optim.metrics.RawPredictions',  # removed in 0.3.5
    ]
    if fullname in removed_cases:
        raise AssertionError(f"class {repr(fullname)} was deprecated and removed in a previous version")
    refactor_cases = [
        ('thelper.modules', 'thelper.nn'),
        ('thelper.samplers', 'thelper.data.samplers'),
        ('thelper.optim.BinaryAccuracy', 'thelper.optim.metrics.Accuracy'),
        ('thelper.optim.CategoryAccuracy', 'thelper.optim.metrics.Accuracy'),
        ('thelper.optim.ClassifLogger', 'thelper.train.utils.ClassifLogger'),
        ('thelper.optim.ClassifReport', 'thelper.train.utils.ClassifReport'),
        ('thelper.optim.ConfusionMatrix', 'thelper.train.utils.ConfusionMatrix'),
        ('thelper.optim.metrics.BinaryAccuracy', 'thelper.optim.metrics.Accuracy'),
        ('thelper.optim.metrics.CategoryAccuracy', 'thelper.optim.metrics.Accuracy'),
        ('thelper.optim.metrics.ClassifLogger', 'thelper.train.utils.ClassifLogger'),
        ('thelper.optim.metrics.ClassifReport', 'thelper.train.utils.ClassifReport'),
        ('thelper.optim.metrics.ConfusionMatrix', 'thelper.train.utils.ConfusionMatrix'),
        ('thelper.transforms.ImageTransformWrapper', 'thelper.transforms.TransformWrapper'),
        ('thelper.transforms.wrappers.ImageTransformWrapper', 'thelper.transforms.wrappers.TransformWrapper'),
    ]
    old_name = fullname
    for old, new in refactor_cases:
        fullname = fullname.replace(old, new)
    if old_name != fullname:
        logger.warning(f"class fullname '{str(old_name)}' was resolved to '{str(fullname)}'")
    return fullname


def import_class(fullname):
    # type: (str) -> Type
    """General-purpose runtime class importer.

    Supported syntax:
        1. ``module.package.Class`` will import the fully qualified ``Class`` located
           in ``package`` from the *installed* ``module``
        2. ``/some/path/mod.pkg.Cls`` will import ``Cls`` as fully qualified ``mod.pkg.Cls`` from
           ``/some/path`` directory

    Args:
        fullname: the fully qualified class name to be imported.

    Returns:
        The imported class.
    """
    assert isinstance(fullname, str)
    fullname = pathlib.Path(fullname).as_posix()
    if "/" in fullname:
        mod_path, mod_cls_name = fullname.rsplit("/", 1)
        pkg_name = mod_cls_name.rsplit(".", 1)[0]
        pkg_file = os.path.join(mod_path, pkg_name.replace(".", "/")) + ".py"
        mod_cls_name = resolve_import(mod_cls_name)
        spec = importlib.util.spec_from_file_location(mod_cls_name, pkg_file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        class_name = mod_cls_name.rsplit('.', 1)[-1]
    else:
        fullname = resolve_import(fullname)
        module_name, class_name = fullname.rsplit('.', 1)
        module = importlib.import_module(module_name)
    return getattr(module, class_name)


def import_function(fullname, params=None):
    # type: (str, Optional[thelper.typedefs.ConfigDict]) -> FunctionType
    """General-purpose runtime function importer, with support for param binding.

    Args:
        fullname: the fully qualified function name to be imported.
        params: optional params dictionary to bind to the function call via functools.

    Returns:
        The imported function, with optionally bound parameters.
    """
    func = import_class(fullname)
    if params is not None:
        if not isinstance(params, dict):
            raise AssertionError("unexpected params dict type")
        return functools.partial(func, **params)
    return func


def check_func_signature(func,      # type: FunctionType
                         params     # type: List[str]
                         ):         # type: (...) -> None
    """Checks whether the signature of a function matches the expected parameter list."""
    if func is None or not callable(func):
        raise AssertionError("invalid function object")
    if params is not None:
        if not isinstance(params, list) or not all([isinstance(p, str) for p in params]):
            raise AssertionError("unexpected param name list format")
        import inspect
        func_sig = inspect.signature(func)
        for p in params:
            if p not in func_sig.parameters:
                raise AssertionError("function missing parameter '%s'" % p)


def encode_data(data, approach="lz4", **kwargs):
    """Encodes a numpy array using a given coding approach.

    Args:
        data: the numpy array to encode.
        approach: the encoding; supports `none`, `lz4`, `jpg`, `png`.

    .. seealso::
        | :func:`thelper.utils.decode_data`
    """
    supported_approaches = ["none", "lz4", "jpg", "png"]
    if approach not in supported_approaches:
        raise AssertionError(f"unexpected approach type (got '{approach}')")
    if approach == "none":
        return data
    elif approach == "lz4":
        import lz4
        return lz4.frame.compress(data, **kwargs)
    elif approach == "jpg" or approach == "jpeg":
        import cv2 as cv
        ret, buf = cv.imencode(".jpg", data, **kwargs)
    elif approach == "png":
        import cv2 as cv
        ret, buf = cv.imencode(".png", data, **kwargs)
    else:
        raise NotImplementedError
    if not ret:
        raise AssertionError("failed to encode data")
    return buf


def decode_data(data, approach="lz4", **kwargs):
    """Decodes a binary array using a given coding approach.

    Args:
        data: the binary array to decode.
        approach: the encoding; supports `none`, `lz4`, `jpg`, `png`.

    .. seealso::
        | :func:`thelper.utils.encode_data`
    """
    supported_approach_types = ["none", "lz4", "jpg", "png"]
    if approach not in supported_approach_types:
        raise AssertionError(f"unexpected approach type (got '{approach}')")
    if approach == "none":
        return data
    elif approach == "lz4":
        import lz4
        return lz4.frame.decompress(data, **kwargs)
    elif approach in ["jpg", "jpeg", "png"]:
        kwargs = copy.deepcopy(kwargs)
        if isinstance(kwargs["flags"], str):  # required arg by opencv
            kwargs["flags"] = eval(kwargs["flags"])
        import cv2 as cv
        return cv.imdecode(data, **kwargs)
    else:
        raise NotImplementedError


def get_class_logger(skip=0):
    """Shorthand to get logger for current class frame."""
    return logging.getLogger(get_caller_name(skip + 1).rsplit(".", 1)[0])


def get_func_logger(skip=0):
    """Shorthand to get logger for current function frame."""
    return logging.getLogger(get_caller_name(skip + 1))


def get_caller_name(skip=2):
    # source: https://gist.github.com/techtonik/2151727
    """Returns the name of a caller in the format module.class.method.

    Args:
        skip: specifies how many levels of stack to skip while getting the caller.

    Returns:
        An empty string is returned if skipped levels exceed stack height; otherwise,
        returns the requested caller name.
    """

    def stack_(frame):
        frame_list = []
        while frame:
            frame_list.append(frame)
            frame = frame.f_back
        return frame_list

    # noinspection PyProtectedMember
    stack = stack_(sys._getframe(1))
    start = 0 + skip
    if len(stack) < start + 1:
        return ""
    parent_frame = stack[start]
    name = []
    module = inspect.getmodule(parent_frame)
    # `modname` can be None when frame is executed directly in console
    if module:
        name.append(module.__name__)
    # detect class name
    if "self" in parent_frame.f_locals:
        # I don't know any way to detect call from the object method
        # XXX: there seems to be no way to detect static method call - it will
        #      be just a function call
        name.append(parent_frame.f_locals["self"].__class__.__name__)
    codename = parent_frame.f_code.co_name
    if codename != "<module>":  # top level usually
        name.append(codename)  # function or a method
    del parent_frame
    return ".".join(name)


def get_key(key, config, msg=None, delete=False):
    """Returns a value given a dictionary key, throwing if not available."""
    if isinstance(key, list):
        if len(key) <= 1:
            if msg is not None:
                raise AssertionError(msg)
            else:
                raise AssertionError("must provide at least two valid keys to test")
        for k in key:
            if k in config:
                val = config[k]
                if delete:
                    del config[k]
                return val
        if msg is not None:
            raise AssertionError(msg)
        else:
            raise AssertionError("config dictionary missing a field named as one of '%s'" % str(key))
    else:
        if key not in config:
            if msg is not None:
                raise AssertionError(msg)
            else:
                raise AssertionError("config dictionary missing '%s' field" % key)
        else:
            val = config[key]
            if delete:
                del config[key]
            return val


def get_key_def(key, config, default=None, msg=None, delete=False):
    """Returns a value given a dictionary key, or the default value if it cannot be found."""
    if isinstance(key, list):
        if len(key) <= 1:
            if msg is not None:
                raise AssertionError(msg)
            else:
                raise AssertionError("must provide at least two valid keys to test")
        for k in key:
            if k in config:
                val = config[k]
                if delete:
                    del config[k]
                return val
        return default
    else:
        if key not in config:
            return default
        else:
            val = config[key]
            if delete:
                del config[key]
            return val


def get_log_stamp():
    """Returns a print-friendly and filename-friendly identification string containing platform and time."""
    return str(platform.node()) + "-" + time.strftime("%Y%m%d-%H%M%S")


def get_git_stamp():
    """Returns a print-friendly SHA signature for the framework's underlying git repository (if found)."""
    try:
        import git
        try:
            repo = git.Repo(path=os.path.abspath(__file__), search_parent_directories=True)
            sha = repo.head.object.hexsha
            return str(sha)
        except (AttributeError, git.InvalidGitRepositoryError):
            return "unknown"
    except (ImportError, AttributeError):
        return "unknown"


def get_env_list():
    """Returns a list of all packages installed in the current environment.

    If the required packages cannot be imported, the returned list will be empty. Note that some
    packages may not be properly detected by this approach, and it is pretty hacky, so use it with
    a grain of salt (i.e. logging is fine).
    """
    try:
        import pip
        # noinspection PyUnresolvedReferences
        pkgs = pip.get_installed_distributions()
        return sorted(["%s %s" % (pkg.key, pkg.version) for pkg in pkgs])
    except (ImportError, AttributeError):
        try:
            import pkg_resources as pkgr
            return sorted([str(pkg) for pkg in pkgr.working_set])
        except (ImportError, AttributeError):
            return []


def str2size(input_str):
    """Returns a (WIDTH, HEIGHT) integer size tuple from a string formatted as 'WxH'."""
    if not isinstance(input_str, str):
        raise AssertionError("unexpected input type")
    display_size_str = input_str.split('x')
    if len(display_size_str) != 2:
        raise AssertionError("bad size string formatting")
    return tuple([max(int(substr), 1) for substr in display_size_str])


def str2bool(s):
    """Converts a string to a boolean.

    If the lower case version of the provided string matches any of 'true', '1', or
    'yes', then the function returns ``True``.
    """
    if isinstance(s, bool):
        return s
    if isinstance(s, (int, float)):
        return s != 0
    if isinstance(s, str):
        positive_flags = ["true", "1", "yes"]
        return s.lower() in positive_flags
    raise AssertionError("unrecognized input type")


def clipstr(s, size, fill=" "):
    """Clips a string to a specific length, with an optional fill character."""
    if len(s) > size:
        s = s[:size]
    if len(s) < size:
        s = fill * (size - len(s)) + s
    return s


def lreplace(string, old_prefix, new_prefix):
    """Replaces a single occurrence of `old_prefix` in the given string by `new_prefix`."""
    return re.sub(r'^(?:%s)+' % re.escape(old_prefix), lambda m: new_prefix * (m.end() // len(old_prefix)), string)


def query_yes_no(question, default=None, bypass=None):
    """Asks the user a yes/no question and returns the answer.

    Args:
        question: the string that is presented to the user.
        default: the presumed answer if the user just hits ``<Enter>``. It must be 'yes',
            'no', or ``None`` (meaning an answer is required).
        bypass: the option to select if the ``bypass_queries`` global variable is set to
            ``True``. Can be ``None``, in which case the function will throw an exception.

    Returns:
        ``True`` for 'yes', or ``False`` for 'no' (or their respective variations).
    """
    valid = {"yes": True, "ye": True, "y": True, "no": False, "n": False}
    if bypass is not None and (not isinstance(bypass, str) or bypass not in valid):
        raise AssertionError("unexpected bypass value")
    clean_q = clipstr(question.replace("\n", " ").replace("\t", " ").replace("'", "`"), 45)
    if bypass_queries:
        if bypass is None:
            raise AssertionError("cannot bypass interactive query, no default value provided")
        logger.debug(f"bypassed query '{clean_q}...' with {valid[bypass]}")
        return valid[bypass]
    if (isinstance(default, bool) and default) or \
       (isinstance(default, str) and default.lower() in ["yes", "ye", "y"]):
        prompt = " [Y/n] "
    elif (isinstance(default, bool) and not default) or \
         (isinstance(default, str) and default.lower() in ["no", "n"]):
        prompt = " [y/N] "
    else:
        prompt = " [y/n] "
    sys.stdout.flush()
    sys.stderr.flush()
    time.sleep(0.25)  # to make sure all debug/info prints are done, and we see the question
    while True:
        sys.stdout.write(question + prompt + "\n>> ")
        choice = input().lower()
        if default is not None and choice == "":
            if isinstance(default, str):
                sys.stdout.write("\n")
                logger.debug(f"defaulted query '{clean_q}...' with {valid[default]}")
                return valid[default]
            else:
                sys.stdout.write("\n")
                logger.debug(f"defaulted query '{clean_q}...' with {default}")
                return default
        elif choice in valid:
            sys.stdout.write("\n")
            logger.debug(f"answered query '{clean_q}...' with {valid[choice]}")
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes/y' or 'no/n'.\n")


def query_string(question, choices=None, default=None, allow_empty=False, bypass=None):
    """Asks the user a question and returns the answer (a generic string).

    Args:
        question: the string that is presented to the user.
        choices: a list of predefined choices that the user can pick from. If
            ``None``, then whatever the user types will be accepted.
        default: the presumed answer if the user just hits ``<Enter>``. If ``None``,
            then an answer is required to continue.
        allow_empty: defines whether an empty answer should be accepted.
        bypass: the returned value if the ``bypass_queries`` global variable is set to
            ``True``. Can be ``None``, in which case the function will throw an exception.

    Returns:
        The string entered by the user.
    """
    clean_q = clipstr(question.replace("\n", " ").replace("\t", " ").replace("'", "`"), 45)
    if bypass_queries:
        if bypass is None:
            raise AssertionError("cannot bypass interactive query, no default value provided")
        logger.debug(f"bypassed query '{clean_q}...' with {bypass}")
        return bypass
    sys.stdout.flush()
    sys.stderr.flush()
    time.sleep(0.25)  # to make sure all debug/info prints are done, and we see the question
    while True:
        msg = question
        if choices is not None:
            msg += "\n\t(choices=%s)" % str(choices)
        if default is not None:
            msg += "\n\t(default=%s)" % default
        sys.stdout.write(msg + "\n>> ")
        answer = input()
        if answer == "":
            if default is not None:
                sys.stdout.write("\n")
                logger.debug(f"defaulted query '{clean_q}...' with {default}")
                return default
            elif allow_empty:
                sys.stdout.write("\n")
                logger.debug(f"answered query '{clean_q}...' with empty string")
                return answer
        elif choices is not None:
            if answer in choices:
                sys.stdout.write("\n")
                logger.debug(f"answered query '{clean_q}...' with choice '{answer}'")
                return answer
        else:
            sys.stdout.write("\n")
            logger.debug(f"answered query '{clean_q}...' with {answer}")
            return answer
        sys.stdout.write("Please respond with a valid string.\n")


def get_config_session_name(config):
    # type: (thelper.typedefs.ConfigDict) -> Optional[str]
    """Returns the 'name' of a session as defined inside a configuration dictionary.

    The current implementation will scan for multiple keywords and return the first value found. If no
    keyword is matched, the function will return None.

    Args:
        config: the configuration dictionary to parse for a name.

    Returns:
        The name that should be given to the session (or 'None' if unknown/unavailable).
    """
    return thelper.utils.get_key_def(["output_dir_name", "output_directory_name",
                                      "session_name", "name"], config, None)


def get_config_output_root(config):
    # type: (thelper.typedefs.ConfigDict) -> Optional[str]
    """Returns the output root directory as defined inside a configuration dictionary.

    The current implementation will scan for multiple keywords and return the first value found. If no
    keyword is matched, the function will return None.

    Args:
        config: the configuration dictionary to parse for a root output directory.

    Returns:
        The path to the output root directory. Can point to a non-existing directory, or be None.
    """
    return thelper.utils.get_key_def(["output_root_dir", "output_root_directory"], config, None)


def get_checkpoint_session_root(ckpt_path):
    # type: (str) -> Optional[str]
    """Returns the session root directory associated with a checkpoint path.

    The given path can point to a checkpoint file or to a directory that contains checkpoints. The
    returned output directory will be the top-level of the session that created the checkpoint, or
    None if it cannot be deduced.

    Args:
        ckpt_path: the path to a checkpoint or to an exisiting directory that contains checkpoints.

    Returns:
        The path to the session root directory. Will always point to an existing directory, or be None.
    """
    assert os.path.exists(ckpt_path), "input path should point to valid filesystem node"
    ckpt_dir_path = os.path.dirname(os.path.abspath(ckpt_path)) \
        if not os.path.isdir(ckpt_path) else os.path.abspath(ckpt_path)
    # find session dir by looking for 'logs' directory
    if os.path.isdir(os.path.join(ckpt_dir_path, "logs")):
        return os.path.abspath(os.path.join(ckpt_dir_path, ".."))
    elif os.path.isdir(os.path.join(ckpt_dir_path, "../logs")):
        return os.path.abspath(os.path.join(ckpt_dir_path, "../.."))
    return None  # cannot be found... giving up


def get_save_dir(out_root, dir_name, config=None, resume=False, backup_ext=".json"):
    """Returns a directory path in which the app can save its data.

    If a folder with name ``dir_name`` already exists in the directory ``out_root``, then the user will be
    asked to pick a new name. If the user refuses, ``sys.exit(1)`` is called. If config is not ``None``, it
    will be saved to the output directory as a json file. Finally, a ``logs`` directory will also be created
    in the output directory for writing logger files.

    Args:
        out_root: path to the directory root where the save directory should be created.
        dir_name: name of the save directory to create. If it already exists, a new one will be requested.
        config: dictionary of app configuration parameters. Used to overwrite i/o queries, and will be
            written to the save directory in json format to test writing. Default is ``None``.
        resume: specifies whether this session is new, or resumed from an older one (in the latter
            case, overwriting is allowed, and the user will never have to choose a new folder)
        backup_ext: extension to use when creating configuration file backups.

    Returns:
        The path to the created save directory for this session.
    """
    func_logger = get_func_logger()
    if config is not None:
        config_out_root = thelper.utils.get_config_output_root(config)
        if out_root is not None and config_out_root is not None and out_root != config_out_root:
            answer = query_string("Received conflicting output root directory paths; which one should be used?\n"
                                  f"\t[config] = {config_out_root}\n\t[cli] = {out_root}", choices=["config", "cli"],
                                  default="config", bypass="config")
            if answer == "config":
                out_root = config_out_root
        elif out_root is None and config_out_root is not None:
            out_root = config_out_root
        config_dir_name = thelper.utils.get_config_session_name(config)
        if config_dir_name is not None:
            assert isinstance(config_dir_name, str), "config session/directory name should be given as string"
            assert not os.path.isabs(config_dir_name), "config session/directory name should never be full (abs) path"
            if dir_name is not None and dir_name != config_dir_name:
                func_logger.warning(f"overriding output session directory name '{dir_name}' to '{config_dir_name}'")
            dir_name = config_dir_name
    if out_root is None:
        time.sleep(0.25)  # to make sure all debug/info prints are done, and we see the question
        out_root = query_string("Please provide the path to where session directories should be created/saved:")
    func_logger.info(f"output root directory = {os.path.abspath(out_root)}")
    os.makedirs(out_root, exist_ok=True)
    save_dir = os.path.join(out_root, dir_name) if dir_name is not None else out_root
    if not resume:
        overwrite = str2bool(config["overwrite"]) if config is not None and "overwrite" in config else False
        time.sleep(0.25)  # to make sure all debug/info prints are done, and we see the question
        while os.path.exists(save_dir) and not overwrite:
            abs_save_dir = os.path.abspath(save_dir).replace("\\", "/")
            overwrite = query_yes_no("Training session at '%s' already exists; overwrite?" % abs_save_dir, bypass="y")
            if not overwrite:
                save_dir = query_string("Please provide a new save directory path:")
    func_logger.info(f"output session directory = {os.path.abspath(save_dir)}")
    os.makedirs(save_dir, exist_ok=True)
    logs_dir = os.path.join(save_dir, "logs")
    func_logger.info(f"output logs directory = {os.path.abspath(logs_dir)}")
    os.makedirs(logs_dir, exist_ok=True)
    if config is not None:
        common_backup_path = os.path.join(save_dir, "config.latest" + backup_ext)
        if resume and os.path.exists(common_backup_path):
            config_backup = thelper.utils.load_config(common_backup_path, add_name_if_missing=False)
            if config_backup != config:  # TODO make config dict comparison smarter...?
                query_msg = f"Config backup in '{common_backup_path}' differs from config loaded through checkpoint; overwrite?"
                answer = query_yes_no(query_msg, bypass="y")
                if answer:
                    func_logger.warning("config mismatch with previous run; "
                                        "will overwrite latest backup in save directory")
                else:
                    func_logger.error("config mismatch with previous run; user aborted")
                    sys.exit(1)
        save_config(config, common_backup_path)
        tagged_backup_path = os.path.join(logs_dir, "config." + thelper.utils.get_log_stamp() + backup_ext)
        save_config(config, tagged_backup_path)
    return save_dir


def load_config(path, as_json=False, add_name_if_missing=True, **kwargs):
    # type: (str, bool, bool, **Any) -> thelper.typedefs.ConfigDict
    """Loads the configuration dictionary from the provided path.

    The type of file that is loaded is based on the extension in the path.

    If the loaded configuration dictionary does not contain a 'name' field, the name of
    the file itself will be inserted as a value.

    Args:
        path: the path specifying which configuration to be loaded.
            only supported types are loaded unless `as_json` is `True`.
        as_json: specifies if an alternate extension should be considered as JSON format.
        add_name_if_missing: specifies whether the file name should be added to the config
            dictionary if it is missing a 'name' field.
        kwargs: forwarded to the extension-specific importer.
    """
    global fixed_yaml_parsing
    if not fixed_yaml_parsing:
        # https://stackoverflow.com/questions/30458977/yaml-loads-5e-6-as-string-and-not-a-number
        loader = yaml.SafeLoader
        loader.add_implicit_resolver(
            u'tag:yaml.org,2002:float',
            re.compile(u'''^(?:
                 [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
                |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
                |\\.[0-9_]+(?:[eE][-+][0-9]+)?
                |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
                |[-+]?\\.(?:inf|Inf|INF)
                |\\.(?:nan|NaN|NAN))$''', re.X),
            list(u'-+0123456789.'))
        fixed_yaml_parsing = True
    ext = os.path.splitext(path)[-1]
    if ext in [".json", ".yml", ".yaml"] or as_json:
        with open(path) as fd:
            assert not kwargs, "yaml safe load takes no extra args"
            config = yaml.safe_load(fd)  # also supports json
    elif ext == ".pkl":
        with open(path, "rb") as fd:
            config = pickle.load(fd, **kwargs)
    else:
        raise AssertionError(f"unknown input file type: {ext}")
    if add_name_if_missing and thelper.utils.get_config_session_name(config) is None:
        config["name"] = os.path.splitext(os.path.basename(path))[0]
    return config


def save_config(config, path, force_convert=True, as_json=False, **kwargs):
    # type: (thelper.typedefs.ConfigDict, str, bool, bool, **Any) -> None
    """Saves the given session/object configuration dictionary to the provided path.

    The type of file that is created is based on the extension specified in the path. If the file
    cannot hold some of the objects within the configuration, they will be converted to strings before
    serialization, unless `force_convert` is set to `False` (in which case the function will raise
    an exception).

    Args:
        config: the session/object configuration dictionary to save.
        path: the path specifying where to create the output file. The extension used will determine
            what type of backup to create (e.g. Pickle = .pkl, JSON = .json, YAML = .yml/.yaml).
            if `as_json` is `True`, then any specified extension will be preserved bump dumped as JSON.
        force_convert: specifies whether non-serializable types should be converted if necessary.
        as_json: specifies if an alternate extension should be considered as JSON format.
        kwargs: forwarded to the extension-specific exporter.
    """
    ext = os.path.splitext(path)[-1]
    if ext in [".json", ".yml", ".yaml"] or as_json:
        with open(path, "w") as fd:
            kwargs.setdefault("indent", 4)
            if ext == ".json" or as_json:
                serializer = (lambda x: str(x)) if force_convert else None
                kwargs.setdefault("default", serializer)
                kwargs.setdefault("sort_keys", False)
                json.dump(config, fd, **kwargs)
            else:
                yaml.dump(config, fd, **kwargs)
    elif ext == ".pkl":
        with open(path, "wb") as fd:
            pickle.dump(config, fd, **kwargs)
    else:
        raise AssertionError(f"unknown output file type: {ext}")


def save_env_list(path):
    """Saves a list of all packages installed in the current environment to a log file.

    Args:
        path: the path where the log file should be created.
    """
    with open(path, "w") as fd:
        pkgs_list = thelper.utils.get_env_list()
        if pkgs_list:
            for pkg in pkgs_list:
                fd.write("%s\n" % pkg)
        else:
            fd.write("<n/a>\n")


def is_scalar(val):
    """Returns whether the input value is a scalar according to numpy and PyTorch."""
    if np.isscalar(val):
        return True
    if isinstance(val, torch.Tensor) and (val.dim() == 0 or val.numel() == 1):
        return True
    return False


def to_numpy(array):
    """Converts a list or PyTorch tensor to numpy. Does nothing if already a numpy array."""
    if isinstance(array, list):
        return np.asarray(array)
    elif isinstance(array, torch.Tensor):
        return array.cpu().numpy()
    elif isinstance(array, np.ndarray):
        return array
    else:
        raise AssertionError(f"unexpected input type ({type(array)})")


def stringify_confmat(confmat, class_list, hide_zeroes=False, hide_diagonal=False, hide_threshold=None):
    """Transforms a confusion matrix array obtained in list or numpy format into a printable string."""
    if not isinstance(confmat, np.ndarray) or not isinstance(class_list, list):
        raise AssertionError("invalid inputs")
    column_width = 9
    empty_cell = " " * column_width
    fst_empty_cell = (column_width - 3) // 2 * " " + "t/p" + (column_width - 3) // 2 * " "
    if len(fst_empty_cell) < len(empty_cell):
        fst_empty_cell = " " * (len(empty_cell) - len(fst_empty_cell)) + fst_empty_cell
    res = "\t" + fst_empty_cell + " "
    for label in class_list:
        res += ("%{0}s".format(column_width) % clipstr(label, column_width)) + " "
    res += ("%{0}s".format(column_width) % "total") + "\n"
    for idx_true, label in enumerate(class_list):
        res += ("\t%{0}s".format(column_width) % clipstr(label, column_width)) + " "
        for idx_pred, _ in enumerate(class_list):
            cell = "%{0}d".format(column_width) % int(confmat[idx_true, idx_pred])
            if hide_zeroes:
                cell = cell if int(confmat[idx_true, idx_pred]) != 0 else empty_cell
            if hide_diagonal:
                cell = cell if idx_true != idx_pred else empty_cell
            if hide_threshold:
                cell = cell if confmat[idx_true, idx_pred] > hide_threshold else empty_cell
            res += cell + " "
        res += ("%{0}d".format(column_width) % int(confmat[idx_true, :].sum())) + "\n"
    res += ("\t%{0}s".format(column_width) % "total") + " "
    for idx_pred, _ in enumerate(class_list):
        res += ("%{0}d".format(column_width) % int(confmat[:, idx_pred].sum())) + " "
    res += ("%{0}d".format(column_width) % int(confmat.sum())) + "\n"
    return res


def get_glob_paths(input_glob_pattern, can_be_dir=False):
    """Parse a wildcard-compatible file name pattern for valid file paths."""
    glob_file_paths = glob.glob(input_glob_pattern)
    if not glob_file_paths:
        raise AssertionError("invalid input glob pattern '%s' (no matches found)" % input_glob_pattern)
    for file_path in glob_file_paths:
        if not os.path.isfile(file_path) and not (can_be_dir and os.path.isdir(file_path)):
            raise AssertionError("invalid input file at globed path '%s'" % file_path)
    return glob_file_paths


def get_file_paths(input_path, data_root, allow_glob=False, can_be_dir=False):
    """Parse a wildcard-compatible file name pattern at a given root level for valid file paths."""
    if os.path.isabs(input_path):
        if '*' in input_path and allow_glob:
            return get_glob_paths(input_path)
        elif not os.path.isfile(input_path) and not (can_be_dir and os.path.isdir(input_path)):
            raise AssertionError("invalid input file at absolute path '%s'" % input_path)
    else:
        if not os.path.isdir(data_root):
            raise AssertionError("invalid dataset root directory at '%s'" % data_root)
        input_path = os.path.join(data_root, input_path)
        if '*' in input_path and allow_glob:
            return get_glob_paths(input_path)
        elif not os.path.isfile(input_path) and not (can_be_dir and os.path.isdir(input_path)):
            raise AssertionError("invalid input file at path '%s'" % input_path)
    return [input_path]


def get_params_hash(*args, **kwargs):
    """Returns a sha1 hash for the given list of parameters (useful for caching)."""
    # by default, will use the repr of all params but remove the 'at 0x00000000' addresses
    clean_str = re.sub(r" at 0x[a-fA-F\d]+", "", str(args) + str(kwargs))
    return hashlib.sha1(clean_str.encode()).hexdigest()



import copy
import os
import shutil
from typing import Any, Optional  # noqa: F401

import numpy as np
import pytest
import torch

import thelper

test_save_path = ".pytest_cache"

test_classif_mnist_name = "test-classif-mnist"
test_classif_mnist_path = os.path.join(test_save_path, test_classif_mnist_name)
test_classif_mnist_ft_path = os.path.join(test_save_path, test_classif_mnist_name + "-finetune")


@pytest.fixture
def config(request):
    def fin():
        shutil.rmtree(test_classif_mnist_path, ignore_errors=True)
        shutil.rmtree(test_classif_mnist_ft_path, ignore_errors=True)
    fin()
    request.addfinalizer(fin)
    return {
        "name": test_classif_mnist_name,
        "bypass_queries": True,
        "datasets": {
            "mnist": {
                "type": "torchvision.datasets.MNIST",
                "params": {
                    "root": os.path.join(test_classif_mnist_path, "mnist"),
                    "train": False,  # use test set, its smaller (quicker test)
                    "download": True
                },
                "task": {
                    "type": "thelper.tasks.Classification",
                    "params": {
                        "class_names": [
                            "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"
                        ],
                        "input_key": "0",
                        "label_key": "1"
                    }
                }
            }
        },
        "loaders": {
            "shuffle": True,
            "batch_size": 32,
            "workers": 0,
            "skip_class_balancing": True,
            "base_transforms": [
                {
                    "operation": "thelper.transforms.NormalizeMinMax",
                    "params": {
                        "min": [127],
                        "max": [255]
                    }
                },
                {
                    "operation": "thelper.transforms.Unsqueeze",
                    "params": {
                        "axis": 0
                    }
                }

            ],
            "train_scale": 0.1,
            "train_split": {
                "mnist": 0.8
            },
            "valid_scale": 0.1,
            "valid_split": {
                "mnist": 0.1
            },
            "test_scale": 0.1,
            "test_split": {
                "mnist": 0.1
            }
        },
        "model": {
            "type": "thelper.nn.lenet.LeNet"
        },
        "trainer": {
            "epochs": 2,
            "monitor": "accuracy",
            "optimization": {
                "loss": {
                    "type": "torch.nn.CrossEntropyLoss"
                },
                "optimizer": {
                    "type": "torch.optim.Adam",
                    "params": {
                        "lr": 0.001
                    }
                }
            },
            "metrics": {
                "accuracy": {
                    "type": "thelper.optim.Accuracy",
                    "params": {
                        "top_k": 1
                    }
                }
            }
        }
    }


def test_reload(config):
    train_outputs = thelper.cli.create_session(config, test_save_path)
    assert len(train_outputs) == 2
    assert train_outputs[0]["train/metrics"]["accuracy"] < train_outputs[1]["train/metrics"]["accuracy"]
    ckptdata = thelper.utils.load_checkpoint(test_classif_mnist_path, always_load_latest=True)
    override_config = copy.deepcopy(config)
    override_config["trainer"]["epochs"] = 3
    resume_outputs = thelper.cli.resume_session(ckptdata, save_dir=test_save_path, config=override_config)
    assert len(resume_outputs) == 3
    assert train_outputs[1]["train/metrics"]["accuracy"] < resume_outputs[2]["train/metrics"]["accuracy"]
    ckptdata = thelper.utils.load_checkpoint(test_classif_mnist_path)
    eval_outputs = thelper.cli.resume_session(ckptdata, save_dir=test_save_path, eval_only=True)
    assert any(["test/metrics" in v for v in eval_outputs.values()])
    override_config["trainer"]["epochs"] = 1
    override_config["model"] = {"ckptdata": test_classif_mnist_path}
    override_config["name"] += "-finetune"
    finetune_outputs = thelper.cli.create_session(override_config, test_save_path)
    assert len(finetune_outputs) == 1
    assert finetune_outputs[0]["train/metrics"]["accuracy"] > train_outputs[1]["train/metrics"]["accuracy"]


def compare_dictionaries(dictA, dictB, dictA_name="A", dictB_name="B", path=""):
    err, key_err, value_err = "", "", ""
    old_path = path
    for k in dictA.keys():
        path = old_path + "[%s]" % k
        if k not in dictB:
            key_err += "key %s%s not in %s\n" % (dictB_name, path, dictB_name)
        else:
            if isinstance(dictA[k], dict) and isinstance(dictB[k], dict):
                err += compare_dictionaries(dictA[k], dictB[k], dictA_name, dictB_name, path)
            else:
                if dictA[k] != dictB[k]:
                    value_err += "value of %s%s (%s) not same as %s%s (%s)\n"\
                        % (dictA_name, path, dictA[k], dictB_name, path, dictB[k])
    for k in dictB.keys():
        path = old_path + "[%s]" % k
        if k not in dictA:
            key_err += "key %s%s not in %s\n" % (dictB_name, path, dictA_name)
    return key_err + value_err + err


def test_outputs(config):
    override_config = copy.deepcopy(config)
    override_config["trainer"]["use_tbx"] = True
    train_outputs = thelper.cli.create_session(override_config, test_save_path)
    assert len(train_outputs) == 2
    assert train_outputs[0]["train/metrics"]["accuracy"] < train_outputs[1]["train/metrics"]["accuracy"]
    output_path = os.path.join(test_classif_mnist_path, "output", test_classif_mnist_name)
    assert os.path.isdir(output_path)
    out_dirs = next(os.walk(output_path))[1]
    assert len(out_dirs) == 2
    assert any([out_dir.startswith("train-") for out_dir in out_dirs])
    assert any([out_dir.startswith("valid-") for out_dir in out_dirs])
    for out_dir in out_dirs:
        ltype = "train" if out_dir.startswith("train-") else "valid"
        epoch_out_path = os.path.join(output_path, out_dir)
        assert os.path.isdir(epoch_out_path)
        epoch_out_files = next(os.walk(epoch_out_path))[2]
        assert len(epoch_out_files) == 4
        assert "accuracy-0000.txt" in epoch_out_files
        assert "accuracy-0001.txt" in epoch_out_files
        assert "config.json" in epoch_out_files
        assert any([p.startswith("events.out.tfevents.") for p in epoch_out_files])
        for filename in epoch_out_files:
            if filename.startswith("accuracy-"):
                epoch = int(filename.split("-")[1].split(".")[0])
                with open(os.path.join(epoch_out_path, filename), "r") as fd:
                    assert np.isclose(float(fd.readline()), train_outputs[epoch][ltype + "/metrics"]["accuracy"])
            elif filename == "config.json":
                backup_config_path = os.path.join(epoch_out_path, filename)
                backup_config = thelper.utils.load_config(backup_config_path, add_name_if_missing=False)
                assert compare_dictionaries(backup_config, override_config) == ""


def callback(task,          # type: thelper.tasks.utils.Task
             input,         # type: thelper.typedefs.InputType
             pred,          # type: thelper.typedefs.AnyPredictionType
             target,        # type: thelper.typedefs.AnyTargetType
             sample,        # type: thelper.typedefs.SampleType
             loss,          # type: Optional[float]
             iter_idx,      # type: int
             max_iters,     # type: int
             epoch_idx,     # type: int
             max_epochs,    # type: int
             **kwargs,      # type: Any
             ):             # type: (...) -> None
    assert isinstance(task, thelper.tasks.Classification)
    assert isinstance(input, torch.Tensor)
    assert isinstance(pred, torch.Tensor)
    assert isinstance(target, torch.Tensor)
    assert isinstance(sample, dict)
    assert isinstance(iter_idx, int)
    assert isinstance(max_iters, int) and iter_idx < max_iters
    assert isinstance(epoch_idx, int)
    assert isinstance(max_epochs, int) and epoch_idx < max_epochs
    assert "hello" in kwargs
    kwargs["hello"][0] = "bye"


def test_callbacks(config, mocker):
    override_config = copy.deepcopy(config)
    override_config["trainer"]["callback"] = callback
    override_config["trainer"]["callback_kwargs"] = {"hello": ["hi"]}
    override_config["trainer"]["display"] = True
    fake_draw = mocker.patch("thelper.draw.draw")
    assert thelper.cli.create_session(override_config, test_save_path)
    assert fake_draw.call_count > 0
    assert override_config["trainer"]["callback_kwargs"]["hello"][0] == "bye"
